(defpackage #:app-constants
  (:use #:cl
        #:larder.web.config)
    (:local-nicknames (#:config #:larder.web.config))
  (:export #:define-constant
	   #:+false+
	   #:+true+
           ;; ALERT-TYPES
           #:+suc+
           #:+err+
           #:+miss+
           #:+inval+
           #:+cre+
           #:+warn+
           ;; DIRECTORIES
           #:+storage-dir+
           #:+media-dir+
           #:+page-dir+
           #:+recipe-dir+
           #:+snapshot-dir+
           #:+thumb-dir+
           #:+user-dir+
           ;; GENERAL PURPOSE
           #:+valid-uuid-length+
           #:+home-page-title+
           #:+home-page-uuid+
           #:+user-recipe-max-upload-size+))
(in-package #:app-constants)

#| Switched to `DEFINE-CONSTANT' from `DEFCONSTANT'.
================================================================================
Because this website uses Steel Bank Common Lisp (SBCL), I need to go through a
cycle of confirming changes to the constant values even though they have not
changed. This behaviour is explained in the SBCL Manual 2.1.3 2021-03 (Section
2.3.4 Defining Constants, page 5 (printed) page 13 (PDF)). The key part of the
section is,
'ANSI says that doing `DEFCONSTANT' of the same symbol more than once is
undefined unless the new value is eql to the old value.'
http://www.sbcl.org/manual/#Defining-Constants (this URL should provide the
latest information of the subject).
A workaround, provided by the SBCL Manual is to use the `DEFINE-CONSTANT' macro
instead of `DEFCONST'. By doing this, you can use Quickload to reload the code
(after a big change for example) and not have to repeat the cycle of 'updating'
the constants when they have not changed.
|#
(defmacro define-constant (name value &optional doc)
  `(defconstant ,name (if (boundp ',name) (symbol-value ',name) ,value)
     ,@(when doc (list doc))))

#| SQLite does not have Boolean value types.
================================================================================
SQLite, as far as I know, can't store boolean values. So, I have made these
constants to reduce hard-coded `1' and/or `0' values when `TRUE' and
`NIL'/`FALSE' values are want is meant (in the code-base). If you're using
something like MySQL or PostgreSQL, you shouldn't need these. I'm leaving them
here just in case, though.
|#

(define-constant +false+ 0 "An integer representing 'false' (for SQLite mostly).")
(define-constant +true+  1 "An integer representing 'true' (for SQLite mostly).")

#| Alert Type strings when calling `SET-ALERT'.
================================================================================
These constants are just redundancy steps. The `SET-ALERT' function in the
`UTILS' package should not change much going forward. But, the code which calls
this function resides across the code-base. Because of this, there is a lot of
hard-coded string values lying around -- calling the `SET-ALERT' function. I can
see refactoring that code, in the future, as a massive pain in the arse -- if I
decide to change the `ALERT-TYPE' strings the `SET-ALERT' function requires. So,
I've made these constants. This mean I only need to change the values in one
place -- if/when I decided to update the `SET-ALERT' function.
|#
(define-constant +suc+   "success"      "Use when calling `SET-ALERT' function.")
(define-constant +err+   "error"        "Use when calling `SET-ALERT' function.")
(define-constant +miss+  "missing-data" "Use when calling `SET-ALERT' function.")
(define-constant +inval+ "invalid-data" "Use when calling `SET-ALERT' function.")
(define-constant +cre+   "created"      "Use when calling `SET-ALERT' function.")
(define-constant +warn+  "warning"      "Use when calling `SET-ALERT' function.")

#| Storage Directories when dealing with user-generated content
================================================================================
These directories are the standard locations used by the
system. 'user-generated' here means all variations of `ACCOUNT' ('owner',
'admin' and 'user'). Use these constants to reduce typing and reduce the amount
of hard-coded strings out of the code-base.

The use of singular directory names is used to match the naming convention use
in the database for its tables.
|#
(define-constant +storage-dir+  (merge-pathnames #P"storage/"           *application-root*))
(define-constant +media-dir+    (merge-pathnames #P"storage/media/"     *application-root*))
(define-constant +page-dir+     (merge-pathnames #P"storage/page/"      *application-root*))
(define-constant +recipe-dir+   (merge-pathnames #P"storage/recipe/"    *application-root*))
(define-constant +snapshot-dir+ (merge-pathnames #P"snapshot/"          *application-root*))
(define-constant +thumb-dir+    (merge-pathnames #P"storage/thumbnail/" *application-root*))
(define-constant +user-dir+     (merge-pathnames #P"storage/user/"      *application-root*))

#| General Purpose Constants
================================================================================
These are for the constants which don't have a particular section they can be
sorted into.
|#

(define-constant +valid-uuid-length+ 36  "`36' is derived from the `UUID' package.")
(define-constant +home-page-title+ "Home" "The default title for the home page.")
(define-constant +home-page-uuid+ "3c9d0464-8643-5ffe-96e5-ab1769818209"
  "The filename for the home page. This is so the system can create the home page
during the initial setup of the website.")

#| Conversion Notes
================================================================================
 250000 character/bytes (B) =  2 megabits (Mb) 
 625000 character/bytes (B) =  5 megabits (Mb) 
1250000 character/bytes (B) = 10 megabits (Mb) 
1875000 character/bytes (B) = 15 megabits (Mb) 
2500000 character/bytes (B) = 20 megabits (Mb)

To convert bytes (B) to megabits (Mb), multiply by 0.000008. For example,

1250000 * 0.000008 = 10 Mb
3500000 * 0.000008 = 28 Mb
|#
(define-constant +user-recipe-max-upload-size+ 1250000
  "Upload limit for HTML user uploads in bytes (B).
2500000 B = 2 Mb
1250000 B = 10 Mb
The number refers to the length of the (HTML) string posted in a HTTP POST
request (1 character = 1 byte). That string is then compared against
`+USER-RECIPE-MAX-UPLOAD-SIZE+'. The intention here to then either reject or
accept the data, based on it being under or over the 'max-upload-size'. This is
needed here and not enforced at the server (Nginx Apache Etc.) level because
this is a 'user' account type limit. The site still has the means to accept
larger file sizes.")
