(in-package #:cl-user)
(defpackage #:larder.web.config
  (:use #:cl)
  (:import-from #:envy
                #:config-env-var
                #:defconfig)
  (:export #:config
           #:*application-root*
           #:*static-directory*
           #:*template-directory*
           #:appenv
           #:developmentp
           #:productionp
           #:testp
           #:change-app-env
           #:database-name
           #:database-username
           #:database-password))
(in-package #:larder.web.config)

(setf (config-env-var) "APP_ENV")

(defparameter *application-root*   (asdf:system-source-directory :larder.web))
(defparameter *static-directory*   (merge-pathnames #P"static/" *application-root*))
(defparameter *template-directory* (merge-pathnames #P"templates/" *application-root*))

#| NOTE: MAKE SURE DATABASES ARE CREATED BEFORE RUNNING ANY CODE
================================================================================
At the time of writing, the system is using PostgreSQL as its database. Because
of this, PSQL must be installed on the system, with the databases created
beforehand. There should be some SQL scripts in,

`*APPLICATION-ROOT*'/db/schema.sql

for examples on how to create the databases.
You can run (LARDER.WEB.DB:ENSURE-TABLES-EXIST) in SLIME to create the tables
after you've created the database.
|#
(defconfig :common
    ;;`(:databases ((:maindb :sqlite3 :database-name ":memory:"))))
    nil)

;; Use this whilst working on the code-base. Switch over to `|TEST|'
;; config. (see below) when you want to run the system's test suite. Switching
;; to `|TEST|' stops your dev. database from getting clogged up with test data.
(defconfig |development|
    `(:databases ((:maindb
                   :postgres
                   :database-name "larderdev"
                   :username "larder"
                   :password "lwebdb"))
      :debug t))

;; Use this config. when running in production, obviously.
(defconfig |production|
    `(:databases ((:maindb
                   :postgres
                   :database-name "larderprod"
                   :username "larder"
                   :password "lwebdb"))
      :debug nil))

;; Change to this config. when running the systems test suite. The quickest way
;; to do that is to run the following code in SLIME:
;; (defun change-app-env :test t)
;; (asdf:test-system :larder.web)
(defconfig |test|
    `(:databases ((:maindb
                   :postgres
                   :database-name "lardertest"
                   :username "larder"
                   :password "lwebdb"))
      :debug t))

(defun config (&optional key)
  "Returns a CONS list summarising the current environments settings."
  (envy:config #.(package-name *package*) key))

(defun appenv ()
  "Returns a string, stating the current application environment."
  (uiop:getenv (config-env-var #.(package-name *package*))))

(defun developmentp ()
    "Returns a predicate indicating if the app. environment is in DEVELOPMENT."
  (string= (appenv) "development"))

(defun productionp ()
    "Returns a predicate indicating if the app. environment is in PRODUCTION."
  (string= (appenv) "production"))

(defun testp ()
    "Returns a predicate indicating if the app. environment is in TEST."
  (string= (appenv) "test"))

(defun database-name ()
  "Returns the `MAINDB' from the system's config.
At the time of writing, the system only uses one internal database and the name
of that database is what this function returns. If the system decides to use
more than one database in the future, this function might need to be deprecated.
Note: The system, at time of writing, uses Meilisearch and that has its own
database. This function and the config. package don't have access to that part
of the system. It's an integration and not part of this system's code-base."
  (nth 3 (car (config :databases))))

(defun database-username ()
  "Returns the username used to access the database.
In other words, the username created in database's system program, like pgsql
for example."
  (nth 5 (car (config :databases))))

(defun database-password ()
  "Returns the password used to access the database.
In other words, the password used when a user was created using the database's
program, like pgsql."
  (nth 7 (car (config :databases))))

(defun change-app-env (&key dev prod test)
  "Changes the 'APP_ENV' to either `DEV', `PROD' or `TEST'.
'APP_ENV' is an (operating system) ENVIRONMENT VARIABLE.
If more than one argument is used, the order of priority is:
1. dev
2. prod
3. test
For example, if you run (DEFUN CHANGE-APP-ENV (:DEV T :TEST T)), the 'APP_ENV'
variable will be set to `DEV'. If no option is set to `T', the 'APP_ENV' will
default to `DEV'."
  (cond (dev (setf (osicat:environment-variable "APP_ENV") "development"))
        (prod (setf (osicat:environment-variable "APP_ENV") "production"))
        (test (setf (osicat:environment-variable "APP_ENV") "test"))
        (t (format
            *error-output*
            "~&[WARNING] No APP_ENV set. Please set :DEV :PROD or :TEST to T. Defaulting to DEVELOPMENT environment.")
           (setf (osicat:environment-variable "APP_ENV") "development"))))

;; This is here so the `APP_ENV' defaults to something if the `APP_ENV' isn't
;; set upon first lanuach. For example, if you run `(QL:QUICKLOAD :LARDER.WEB)' and start the server straight away. Another example is if the website is ran as a systemd service and the `APP_ENV' is set within the service file. Use the `CHANGE-APP-ENV' function above to change it whilst working on the code-base/SLIME. 
(setf (osicat:environment-variable "APP_ENV") "development")
