(in-package #:cl-user)
(defpackage #:larder.web.db
  (:use #:cl
        ;; larder.web specific packages.
        #:app-constants
        #:models
        #:validation
        #:hermetic
        #:mito
        #:uuid)
  (:import-from #:larder.web.config
                #:config)
  ;; Part of caveman2 project generator. Using mito so datafly not used.
  ;; (#:import-from #:datafly #:*connection*)
  (:import-from #:cl-dbi
                #:connect-cached)
  (:export #:connection-settings
            #:db
            #:with-connection
            ;; SET-UP AND MAINTENANCE
            #:*tables*
            #:drop-all-tables
            #:drop-table
            #:ensure-tables-exist
            #:init-db
            #:migrate-all
            ;; ACCOUNT
            #:create-account
            #:delete-account
            #:get-account
            #:get-all-accounts
            #:set-account-status
            #:update-account
            #:update-account-type
            ;; PAGE
            #:create-page
            #:delete-page
            #:get-all-pages
            #:get-page
            #:get-pages-count
            #:get-paginated-pages
            #:latest-edited-pages
            #:update-page
            ;; RECIPE
            #:create-recipe
            #:delete-recipe
            #:get-all-recipes
            #:get-paginated-recipes
            #:get-recipe
            #:get-recipes
            #:get-recipes-count
            #:get-recipes-with-source
            #:make-all-recipes-private
            #:make-all-recipes-public
            #:make-recipes-with-source-private
            #:make-recipes-with-source-public
            #:update-recipe
            ;; SITE SETTINGS
            #:get-site-settings
            #:nav-menu-slugs
            #:site-settings-exist?
            #:get-system-data
            #:update-show-log-in
            #:update-nav-menu
            #:update-show-search
            #:update-search-url
            #:update-search-index
            #:update-search-api-key
            #:update-sign-up-setting
            #:update-site-name
            #:update-show-site-logo-setting
            ;; USER RECIPE
            #:create-user-recipe
            #:delete-user-recipe
            #:get-all-user-recipes-count
            #:get-paginated-user-recipes
            #:get-user-recipe
            #:get-user-recipes
            #:update-user-recipe
            ;; FILE
            #:create-file
            #:delete-file-entry ; delete-file is built-in function.
            #:get-all-files
            #:get-file
            #:get-files-for-account
            #:get-paginated-files
            #:update-file
            ;; FAVOURITE
            #:create-favourite
            #:delete-favourite
            #:get-favourite-recipes
            #:get-paginated-favourites
            #:is-favourite?
            ;; UNSORTED
            #:delete-row-entry
            #:entry-exists?
            #:get-table-count))
(in-package #:larder.web.db)

(defun connection-settings (&optional (db :maindb))
  (cdr (assoc db (config :databases))))

(defun db (&optional (db :maindb))
  (apply #'connect-cached (connection-settings db)))

(defmacro with-connection (conn &body body)
  `(let ((*connection* ,conn))
     ,@body))

#| larder.web Specific Code
================================================================================
The code above was created using caveman2's project generator. The code below is
specific to the larder.web project.
|#

(defparameter *tables* '(models:account
                         models:favourite
                         models:file
                         models:page
                         models:recipe
                         models:site_settings
                         models:user_recipe)
  "List of the site's DB tables.
This is used mostly in DB migrations and initial setup. The tables are defined
as classes, inheriting from a mito (ORM) class, and can be found in the `MODELS'
package.")

;; SET-UP AND MAINTENANCE
;; =============================================================================
(defun drop-all-tables ()
  "Drops all the tables from the database which are listed in `*TABLES*'."
  (with-connection (db)
    (mapcar (lambda (table)
              (mito:execute-sql (sxql:drop-table table :if-exists t)))
              *tables*)))

(defun drop-table (table-name)
  "Drops `TABLE-NAME' from database."
  (with-connection (db)
    (mito:execute-sql (sxql:drop-table table-name :if-exists t))))

(defun ensure-tables-exist ()
  "Creates tables in the database.
Make sure you've added the tables name into`*TABLES*'. The tables are built with
`MITO' and the classes used to define the tables are in the `MODELS' package."
  (with-connection (db)
    (mapcar #'mito:ensure-table-exists *tables*)))

(defun init-db (site-name username display-name password allow-sign-up
                show-site-logo search-url search-index search-api-key)
  (create-account username display-name password :owner t)
  (with-connection (db)
    (mito:create-dao 'models:site_settings
                     :site-name site-name
                     :show-site-logo show-site-logo
                     :allow-sign-up allow-sign-up
                     :show-log-in t
                     :search-url search-url
                     :search-index search-index
                     :search-api-key search-api-key
                     :show-search t)))

(defun migrate-all ()
  "Migrate the tables after changing any class definition.
The classes should be defined in the `MODELS' package. Note, the classes are
derived from a `MITO' class."
  (with-connection (db)
    (ensure-tables-exist)
    (mapcar #'mito:migrate-table *tables*)))

;; ACCOUNT
;; =============================================================================

(defun create-account (username display-name password &key owner admin user)
  "Creates an account and adds it to the `ACCOUNT' table in the database.

`ADMIN', `OWNER' and `USER' refer to the `ACCOUNT-TYPE'. If more than argument
is set, the type with the highest privileges is used. For example, if the
function is called like this,

(create-account ... :admin t :user t)

The account will be set to 'admin' because that is the higher ranking type of
the two specified here. `OWNER' is the highest, `ADMIN' is the second and `USER'
is the lowest.

If not account-type argument is given, the account defaults to `USER'."
  (with-connection (db)
    (mito:create-dao
     'models:account
     :account-id
     (format nil "~a"
             (uuid:make-v5-uuid uuid:+namespace-url+ username))
     :username username
     :display-name display-name
     :password (hermetic::hash password)
     :active t
     :account-type
     (cond (owner "owner")
           (admin "admin")
           (user "user")
           (t (format
               *error-output*
               "~&[WARNING] Account type was not specified, defaulting to USER.")
              "user")))))

(defun delete-account (account-id)
  "Deletes `ACCOUNT' from the database."
  (with-connection (db)
    (mito:delete-by-values 'models:account :account-id account-id)))

(defun get-account (&key account-id username)
  "Returns an `ACCOUNT' with either the specified `ACCOUNT-ID' or `USERNAME'."
  (with-connection (db)
    (cond (account-id (mito:find-dao 'models:account :account-id account-id))
          (username (mito:find-dao 'models:account :username username)))))

(defun get-all-accounts ()
  "Returns a list of all `ACCOUNT' entries in the database."
  (with-connection (db)
    (mito:select-dao 'models:account
      (sxql:order-by (:asc :display-name)))))

(defun set-account-status (account-id status)
  "Sets the account with specified `ACCOUNT-ID' in account table in database to `STATUS'.
`STATUS' is a Boolean value. `T' to reactivate and `NIL' to deactivate the
`ACCOUNT'."
  (with-connection (db)
    (let ((account-to-update
            (mito:find-dao 'models:account :account-id account-id)))
      (setf (models::account-active account-to-update) status)
      (mito:save-dao account-to-update))))

(defun update-account (account-id &key display-name new-password)
  "Updates `ACCOUNT' in database."
  (with-connection (db)
    (let ((account-to-update
	    (mito:find-dao 'models:account :account-id account-id)))
      (if (not (validation:string-is-nil-or-empty? display-name))
	  (setf (models::account-display-name account-to-update) display-name))
      (if (not (validation:string-is-nil-or-empty? new-password))
	  (setf (models::account-password account-to-update)
                (hermetic::hash new-password)))
      (mito:save-dao account-to-update))))

(defun update-account-type (account-id &key owner admin user)
  "Set `ACCOUNT' to 'owner', 'administrator' or 'user' with `ACCOUNT-ID' in 'account' table."
  (with-connection (db)
    (let ((account-to-update
            (mito:find-dao 'models:account :account-id account-id)))
            (cond (owner (setf (models::account-account-type account-to-update) "owner"))
                  (admin
                   (setf (models::account-account-type account-to-update) "administrator"))
                  (user  (setf (models::account-account-type account-to-update) "user"))
                  (t
                   (format *error-output*
                           "[WARNING] Account type not provided. Account not updated.")))
      (mito:save-dao account-to-update))))

;; PAGE
;; =============================================================================

(defun create-page (title account-id page-id in-nav-menu can-delete is-public)
  "Add a new row to the 'page' table in the the database."
  (with-connection (db)
    (mito:create-dao 'models:page :title title
                                  :owner account-id
                                  :page-id page-id
                                  :enable-nav-menu (if (string= "on"  in-nav-menu) t nil)
                                  :can-delete (if can-delete t nil)
                                  :public (if (string= "on" is-public) t nil))))

(defun delete-page (&key slug page-id)
  "Delete entry in 'page' table in the database."
  (with-connection (db)
    (if slug
        (mito:delete-by-values 'models:page :slug slug)
        (mito:delete-by-values 'models:page :page-id page-id))))

(defun get-all-pages (&key only-public only-private)
  "Returns a list of all `PAGE' entries in the database.
Pass `:ONLY-PUBLIC' as `T' to retrieve only the public entries from the
database. Pass `:ONLY-PRIVATE' as `T' for the private entries. Passing both
arguments at the same time cancels each other other out, so don't do that. Don't
pass any arguments to retrieve all the entries from the database."
  (with-connection (db)
    (mito:select-dao 'models:page
      (if only-private (sxql:where (:= :public "false")))
      (if only-public (sxql:where (:= :public "true")))
      (sxql:order-by (:asc :title)))))

(defun get-page (&key title page-id)
  "Returns a `PAGE' from the database."
  (with-connection (db)
    (if title
        (mito:find-dao 'models:page :title title)
        (mito:find-dao 'models:page :page-id page-id))))

(defun get-pages-count (&key only-public only-private)
  "Returns the total number of entries in page."
  (with-connection (db)
    (cond (only-public (mito:count-dao 'models:page :public "true"))
          (only-private (mito:count-dao 'models:page :public "false"))
          (t (mito:count-dao 'models:recipe)))))

(defun  get-paginated-pages (page page-amount order &key only-public only-private)
  "Returns a list of entries from page table with an offset."
  (with-connection (db)
    (mito:select-dao 'models:page
      (if only-private (sxql:where (:= :public "false")))
      (if only-public (sxql:where (:= :public "true")))
      (sxql:order-by (:desc order))
      (sxql:offset (* (- page 1) page-amount))
      (sxql:limit page-amount))))

(defun latest-edited-pages (amount &optional reverse)
  "Gets the latest `AMOUNT' of edited entries from the database.
`REVERSE' is an optional parameter which puts the most recently
edited article entry as the first item in the list.."
  (with-connection (db)
    (mito:select-dao 'models:page
      (sxql:limit amount)
      (if reverse
	  (sxql:order-by (:desc 'models::page-updated-at))
	  (sxql:order-by 'models::page-updated-at)))))

(defun update-page (page-id &key title is-public in-nav-menu)
  "Updated entry in 'page' table with `PAGE-ID'."
  (with-connection (db)
    (let ((page-to-update (mito:find-dao 'models:page :page-id page-id)))
      (if title (setf (models::page-title page-to-update) title))
      (if is-public
          (setf (models::page-public page-to-update) t)
          (setf (models::page-public page-to-update) nil))
      (if in-nav-menu
          (setf (models::page-enable-nav-menu page-to-update) t)
          (setf (models::page-enable-nav-menu page-to-update) nil))
      (mito:save-dao page-to-update))))


;; RECIPE
;; =============================================================================

(defun create-recipe (account-id title recipe-id source ingredients thumbnail-id)
  "Adds an entry to the 'user_recipe' table in the database."
  (with-connection (db)
    (mito:create-dao 'models:recipe
                     :author-id account-id
                     :source source
                     :public t ; 'Site' recipes are public by default.
                     :recipe-id recipe-id
                     :ingredients ingredients
                     :thumbnail-id thumbnail-id
                     :title title)))

(defun delete-recipe (recipe-id)
  "Deletes entry from 'recipe' table in the database with specified `RECIPE-ID'."
  (with-connection (db) (mito:delete-by-values 'models:recipe :recipe-id recipe-id)))

(defun get-all-recipes (&key only-public only-private)
  "Retrieves a list of all `RECIPE' entries in the 'recipe' table.
Pass `:ONLY-PUBLIC' as `T' to retrieve only the public entries from the
database. Pass `:ONLY-PRIVATE' as `T' for the private entries. Passing both
arguments at the same time cancels each other other out, so don't do that. Don't
pass any arguments to retrieve all the entries from the database."
  (with-connection (db)
    (mito:select-dao 'models:recipe
      (if only-private (sxql:where (:= :public "false")))
      (if only-public (sxql:where (:= :public "true")))
      (sxql:order-by (:asc :updated_at)))))

(defun get-paginated-recipes (page page-amount order &key only-public only-private)
  "Returns a list of recipes with an offset , defined by `PAGE' & `PAGE-AMOUNT'."
  (with-connection (db)
    (mito:select-dao 'models:recipe
      (when only-private (sxql:where (:= :public "false")))
      (when only-public (sxql:where (:= :public "true")))
      (sxql:order-by (:desc order))
      (sxql:offset (* (- page 1) page-amount))
      (sxql:limit page-amount))))

(defun get-recipe (&key recipe-id thumbnail-id)
  "Returns the entry from the 'recipe' table with `RECIPE-ID'."
  (with-connection (db)
    (cond (recipe-id (mito:find-dao 'models:recipe :recipe-id recipe-id))
          (thumbnail-id (mito:find-dao 'models:recipe :thumbnail-id thumbnail-id))
          (t nil))))

(defun get-recipes (&key order amount reverse)
  "Returns a list of entries from recipe table.
`ORDER' specifies the tables column the entries are sorted by, with 'created_at'
or 'updated_at' being the most likely sort-by orders. `AMOUNT' is the number of
entries you want to be retrieved from the database. `REVERSE' is a Boolean value
and indicates if you want the entries to be sorted in descending order or not."
  (with-connection (db)
    (mito:select-dao 'models:recipe
      (if reverse
          (sxql:order-by (:desc order))
          (sxql:order-by order))
      (sxql:limit amount))))

(defun get-recipes-count (&key only-public only-private)
  "Returns the total number of entries in `TABLE-NAME'."
  (with-connection (db)
    (cond (only-public (mito:count-dao 'models:recipe :public "true"))
          (only-private (mito:count-dao 'models:recipe :public "false"))
          (t (mito:count-dao 'models:recipe)))))

(defun get-recipes-with-source (source-url)
  "Returns collection of recipes with specified `SOURCE-URL' from 'recipe' table."
  (with-connection (db)
    (mito:select-dao 'models:recipe
      (if (null source-url)
          (sxql:where (:is-null :source))
          (sxql:where (:like :source source-url))))))

(defun make-all-recipes-private ()
  "Sets the 'public' value to false for all entries in 'recipe' table."
  (with-connection (db)
    (mito:execute-sql (sxql:update 'models:recipe (sxql:set= :public "false")))))

(defun make-all-recipes-public ()
  "Set the 'public' value to true for all entries in 'recipe' table."
  (with-connection (db)
    (mito:execute-sql (sxql:update 'models:recipe (sxql:set= :public "true")))))

(defun make-recipes-with-source-private (source-url)
  "Sets all the recipes, with specified `SOURCE', 'public' value to false."
  (with-connection (db)
    (mito:execute-sql
     (sxql:update 'models:recipe
       (sxql:set= :public "false")
       (if (null source-url)
           (sxql:where (:is-null :source))
           (sxql:where (:like :source source-url)))))))

(defun make-recipes-with-source-public (source-url)
  "Sets all the recipes, with specified `SOURCE', 'public' value to true."
  (with-connection (db)
    (mito:execute-sql
     (sxql:update 'models:recipe
       (sxql:set= :public "true")
       (if (null source-url)
           (sxql:where (:is-null :source))
           (sxql:where (:like :source source-url)))))))

(defun update-recipe (recipe-id &key title is-public)
  "Update entry, with `RECIPE-ID' in 'recipe' table."
   (with-connection (db)
     (let ((recipe-to-update (mito:find-dao 'models:recipe :recipe-id recipe-id)))
       (when title (setf (models::recipe-title  recipe-to-update) title))
       (if is-public
           (setf (models::recipe-public recipe-to-update) t)
           (setf (models::recipe-public recipe-to-update) nil))
       (mito:save-dao recipe-to-update))))

;; SITE SETTINGS
;; =============================================================================

(defun get-site-settings ()
  "Retrieves the data from the `SITE_SETTINGS' table in the database.
You should really find only one entry in this table. If you find yourself with
multiple entries, either something has gone wrong or you're doing something more
advanced beyond the usual scope of this project."
  (with-connection (db)
    (mito:find-dao 'models:site_settings)))

(defun nav-menu-slugs ()
  (with-connection (db)
    (mito:select-dao 'page
      (sxql:where (:= :enable-nav-menu +true+)))))

(defun site-settings-exist? ()
  "Checks to see if the database contains any `SITE_SETTINGS' table.
Returns `T' if database has at least 1 entry in the `SITE_SETTINGS' table and
`NIL' if no entries are found. You'll tend to use this as part of the 'first
run/setup' checks. If no entries are found, it probably means the site has just
been created and needs to be set up."
  (if (= 0 (with-connection (db) (mito:count-dao 'models:site_settings))) nil t))

(defun get-system-data ()
  "Gets the website's settings and nav-menu slugs from the database."
  (list (get-site-settings) (nav-menu-slugs)))

(defun update-nav-menu (selected-pages)
  "Updates the 'enable_nav_menu' property in 'page' table of the database."
  (loop for page in selected-pages
	do (with-connection (db)
	     (let ((page-to-update (mito:find-dao 'models:page :slug (car page))))
               (if (string= "on" (cdr page))
	           (setf (models::page-enable-nav-menu page-to-update) t)
                   (setf (models::page-enable-nav-menu page-to-update) nil))
	       (mito:save-dao page-to-update)))))

(defun update-show-log-in (show-log-in)
  "Sets the 'show-log-in' setting in 'site_settings' table in the database."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (if (string= "on" show-log-in)
          (setf (models::site_settings-show-log-in settings-to-update) t)
          (setf (models::site_settings-show-log-in settings-to-update) nil))
      (mito:save-dao settings-to-update))))

(defun update-show-search (show-search)
  "Updates the 'show-search'' property in the 'site_settings' table."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (if (string= "on" show-search)
          (setf (models::site_settings-show-search settings-to-update) t)
          (setf (models::site_settings-show-search settings-to-update) nil))
      (mito:save-dao settings-to-update))))
          

(defun update-search-url (search-url)
  "Updates `SEARCH-URL' column in the 'site_settings' table in the database."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (setf (models::site_settings-search-url settings-to-update) search-url)
      (mito:save-dao settings-to-update))))

(defun update-search-index (search-index)
  "Updates `SEARCH-INDEX' column in the 'site_settings' table in the database."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (setf (models::site_settings-search-index settings-to-update) search-index)
      (mito:save-dao settings-to-update))))

(defun update-search-api-key (api-key)
  "Updates the 'search_api_key' column in 'site_settings' table in the database."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (setf (models::site_settings-search-api-key settings-to-update) api-key)
      (mito:save-dao settings-to-update))))

(defun update-sign-up-setting (value)
  "Updates the 'Enable Sign Up' setting in the database with `VALUE'."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (if (string= "on" value)
          (setf (models::site_settings-allow-sign-up settings-to-update) t)
          (setf (models::site_settings-allow-sign-up settings-to-update) nil))
      (mito:save-dao settings-to-update))))

(defun update-site-name (name)
  "Updates the website's `SITE-NAME' the database."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (setf (models::site_settings-site-name settings-to-update) name)
      (mito:save-dao settings-to-update))))

(defun update-show-site-logo-setting (value)
  "Updates the 'Show Site Logo' setting in the database with `VALUE'."
  (with-connection (db)
    (let ((settings-to-update (mito:find-dao 'models:site_settings)))
      (if (string= "on" value)
          (setf (models::site_settings-show-site-logo settings-to-update) t)
          (setf (models::site_settings-show-site-logo settings-to-update) nil))
      (mito:save-dao settings-to-update))))

;; USER RECIPE
;; =============================================================================

(defun create-user-recipe (account-id title is-public user-recipe-id)
  "Adds an entry to the 'user_recipe' table in the database."
  (with-connection (db)
    (mito:create-dao 'models:user_recipe
                     :account-id account-id
                     :public (if (string= "on" is-public) t nil)
                     :title title
                     :user-recipe-id user-recipe-id)))

(defun delete-user-recipe (user-recipe-id)
  "Deletes `USER-RECIPE' entry from the 'user_recipe' table, with `USER-RECIPE-ID'."
  (with-connection (db)
    (mito:delete-by-values 'models:user_recipe :user-recipe-id user-recipe-id)))

(defun get-paginated-user-recipes
    (page page-amount order account-id &key only-public only-private)
  "Returns a list of entries from user_recipe table with an offset."
  (with-connection (db)
    (mito:select-dao 'models:user_recipe
      (sxql:where (:= :account-id account-id))
      (when only-private (sxql:where (:= :public "false")))
      (when only-public (sxql:where (:= :public "true")))
      (sxql:order-by (:desc order))
      (sxql:offset (* (- page 1) page-amount))
      (sxql:limit page-amount))))

(defun get-all-user-recipes-count (&key only-public only-private)
  "Returns the total number of entries in user_recipe table."
  (with-connection (db)
    (cond (only-public (mito:count-dao 'models:user_recipe :public "true"))
          (only-private (mito:count-dao 'models:user_recipe :public "false"))
          (t (mito:count-dao 'models:recipe)))))

(defun get-user-recipe (user-recipe-id)
  "Returns the entry from the 'user_recipe' table with `USER-RECIPE-ID'."
  (with-connection (db)
    (mito:find-dao 'models:user_recipe :user-recipe-id user-recipe-id)))

(defun get-user-recipes (account-id)
  "Gets all the entries in 'user_recipe' table for `ACCOUNT' with `ACCOUNT-ID'."
  (with-connection (db)
    (mito:select-dao 'models:user_recipe
      (sxql:where (:= :account-id account-id))
      (sxql:order-by (:asc :updated_at)))))

(defun update-user-recipe (user-recipe-id &key title is-public)
  "Update entry, with `USER-RECIPE-ID' in 'user_recipe' table."
  (with-connection (db)
    (let ((recipe-to-update
            (mito:find-dao 'models:user_recipe :user-recipe-id user-recipe-id)))
      (when title
          (setf (models::user_recipe-title  recipe-to-update) title))
      (if is-public
          (setf (models::user_recipe-public recipe-to-update) t)
          (setf (models::user_recipe-public recipe-to-update) nil))
      (mito:save-dao recipe-to-update))))

;; FILE
;; =============================================================================

(defun create-file (title mime account-id thumbnail-id is-public file-id)
  "Adds an entry to the 'file' table in the database."
  (with-connection (db)
    (mito:create-dao 'models:file
                     :title title
                     :mime mime
                     :account-id account-id
                     :thumbnail-id thumbnail-id
                     :public (if is-public t nil)
                     :file-id file-id)))

(defun delete-file-entry (file-id)
  "Delete entry from 'file' table in database, with specified `FILE-ID'..

`DELETE-FILE' is a built-in function which deals with files on disk. That is why
`DELETE-FILE-ENTRY' is used here and breaks, the somewhat almost, standard
naming convention in this package.

Note, image files with thumbnails do not have their thumbnail files stored in
the database as a separate entry. The `THUMBNAIL-ID' is included in the main
image entry. So, yo don't need to call this function twice when dealing with
image file entries (in the database)."
  (with-connection (db) (mito:delete-by-values 'models:file :file-id file-id)))

(defun get-file (file-id)
  "Returns entry from 'file' table with `FILE-ID'."
  (with-connection (db)
    (mito:find-dao 'models:file :file-id file-id)))

(defun get-all-files ()
  "Returns all entries in 'file' table."
  (with-connection (db) (mito:select-dao 'models:file)))

(defun get-files-for-account (account-id &key amount order)
  "Returns entries from 'file' table for `ACCOUNT-ID'.
Use `AMOUNT' and `ORDER' to set the number of entries you want to return and the
order you want them returned in. Not specifying and `AMOUNT', returns all the
entries from the database for with the `ACCOUNT-ID'."
  (with-connection (db)
    (mito:select-dao 'models:file
      (sxql:where (:= :account-id account-id))
      (if amount (sxql:limit amount))
      (if order (sxql:order-by order)))))

(defun get-paginated-files (page page-amount order)
  "Returns a list of recipes with an offset , defined by `PAGE' & `PAGE-AMOUNT'."
  (with-connection (db)
    (mito:select-dao 'models:file
      (sxql:order-by order)
      (sxql:offset (* (- page 1) page-amount))
      (sxql:limit page-amount))))

(defun update-file (file-id &key title is-public)
  "Updates entry in 'file' table with `FILE-ID'."
  (with-connection (db)
    (let ((file-to-update (mito:find-dao 'models:file :file-id file-id)))
      (if title (setf (models::file-title file-to-update) title))
      (if is-public ; Is public so set to private.
          (setf (models::file-public file-to-update) nil)
          (setf (models::file-public file-to-update) t))
      (mito:save-dao file-to-update))))

;; FAVOURITE
;; =============================================================================

(defun create-favourite (account-id recipe-id)
  "Adds an entry to the favourite table in the database."
  (with-connection (db)
    (mito:create-dao 'models:favourite
                     :account-id account-id
                     :recipe-id recipe-id)))

(defun delete-favourite (account-id recipe-id)
  "Deletes the row in 'favourite' table with `ACCOUNT-ID' & `RECIPE-ID'."
  (with-connection (db)
    (mito:delete-by-values 'models:favourite
                           :account-id account-id
                           :recipe-id recipe-id)))


(defun get-favourite-recipes (account-id &key limit reverse)
  "Returns a list of 'favourite' entries for `ACCOUNT-ID' from DB.
Use `LIMIT' to specify the maximum number of entries you would like returned
from the database."
  (with-connection (db)
    (mito:select-dao 'models:recipe
      (sxql:where (:= :account-id account-id))
      (when reverse (sxql:order-by (:desc :recipe.created_at)))
      (when limit (sxql:limit limit))
      (sxql:inner-join 'models:favourite
                       :on (:= :favourite.recipe_id :recipe.recipe_id)))))

(defun get-paginated-favourites (page page-amount order account-id)
  "Returns a list of entries from favourite table with an offset."
  (with-connection (db)
    (mito:select-dao 'models:recipe
      (sxql:where (:= :account-id account-id))
      (sxql:order-by (:desc order))
      (sxql:offset (* (- page 1) page-amount))
      (sxql:limit page-amount)
      (sxql:inner-join 'models:favourite
                       :on (:= :favourite.recipe_id :recipe.recipe_id)))))

(defun is-favourite? (account-id recipe-id)
  "Checks to see if 'favourite' entry matching `ACCOUNT-ID' & `RECIPE-ID' exists."
  (with-connection (db)
    (mito:find-dao 'models:favourite
                   :account-id account-id
                   :recipe-id recipe-id)))

#| UNSORTED
================================================================================
Add rough code here whilst you're working stuff out. After that, move it to its
corresponding section above. Make sure you keep the `:EXPORT' section organised,
as well.
|#

(defun delete-row-entry (table-name id-key id-value)
  "Delete a row from `TABLE-NAME' (as a string) in the database.
The `ID-KEY' and `ID-VALUE' parameters are generic values to indentify which row
to delete from the specified table (via `TABLE-NAME'). Both parameters should be
passed as symbols, so use `READ-FROM-STRING' function when source of parameters
are strings (E.G. data posted as a string within a HTML form, via a HTTP POST)."
  (handler-case
      (with-connection (db)
        (mito:delete-by-values table-name id-key id-value))
    (error (e)
      (format *error-output* "~&ERROR: ~a" e)
      nil)))

(defun entry-exists? (table-name id-key id-value)
  "Checks to see if an entry exists in the specified `TABLE-NAME' in the DB.
The `ID-KEY' and `ID-VALUE' parameters are generic values to indentify which row
to check for, from the specified table (via `TABLE-NAME', as string). Both
parameters should be passed as symbols, so use `READ-FROM-STRING' function when
source of parameters are strings (E.G. data posted as a string within a HTML
form, via a HTTP POST)."
  (handler-case
  (with-connection (db)
    (mito:find-dao table-name id-key id-value))
    (error (e)
      (format *error-output* "~&ERROR: ~a" e)
      nil)))

(defun drop-database (db-name)
  ""
  (handler-case
      (with-connection (db)
        (mito:execute-sql (format nil "DROP DATABASE ~a" db-name)))
    (error (e)
      (format *error-output* "[ERROR] ~a~&" e))))

(defun get-table-count (table-name)
  "Returns the total number of entries in `TABLE-NAME'."
  (with-connection (db)
    (mito:count-dao table-name)))
