(in-package #:cl-user)
(defpackage #:meilisearch
  (:use #:cl)
  (:import-from #:dexador
		#:request)
  (:export #:check-health
           #:create-dump
           #:delete-all-entries
           #:delete-db-index
           #:delete-entry
           #:find-entry
           #:get-keys
           #:get-ranking-rules
           #:get-search-attributes
           #:get-stats
           #:set-filter-attributes
           #:set-ranking-rules
           #:set-search-attributes
           #:submit-entry
           #:submit-batch-entries))
(in-package #:meilisearch)

(defun check-health (root-url)
  "Checks the status of the entire Meilisearch database, not individual indexes."
  (handler-case
      (dexador:request
       (format nil "~a/health" root-url)
       :method :get
       :use-connection-pool nil
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun create-dump (root-url meilisearch-api-key)
  "Creates a dump of the entire Meilisearch database."
  (handler-case
      (dexador:request
       (format nil "~a/dumps" root-url)
       :method :post
       :use-connection-pool nil
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun delete-all-entries (root-url meilisearch-index meilisearch-api-key)
  "Deletes all the entries from the Meilisearch database."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/documents" root-url meilisearch-index)
       :method :delete
       :use-connection-pool nil
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun delete-db-index (root-url meilisearch-index meilisearch-api-key)
  "Deletes the specified index from the Meilisearch database, via `MEILISEARCH-INDEX'."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a" root-url meilisearch-index)
       :method :delete
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :use-connection-pool nil
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun delete-entry (id root-url meilisearch-index meilisearch-api-key)
  "Deletes the entry with `ID' from Meilisearch database."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/documents/~a" root-url meilisearch-index id)
       :method :delete
       :use-connection-pool nil
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun get-keys (root-url meilisearch-api-key)
  "Returns a list of API keys used by the Meilisearch instance."
  (handler-case
      (dexador:request
       (format nil "~a/keys" root-url)
       :method :get
       :use-connection-pool nil
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))


(defun find-entry (id root-url meilisearch-index meilisearch-api-key)
  "Finds the entry in the meilisearch database by its `ID'."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/search" root-url meilisearch-index)
       :method :post
       :use-connection-pool nil
       :headers `(("Content-Type" . "application/json")
                  ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :content (format nil "{ \"q\": \"~a\", \"limit\": 1 }" id)
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun get-ranking-rules (root-url meilisearch-index meilisearch-api-key)
  ""
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/settings/ranking-rules" root-url meilisearch-index)
       :method :get
       :use-connection-pool nil
       :headers `(("Content-Type" . "application/json")
                  ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun get-search-attributes (root-url meilisearch-index meilisearch-api-key)
  ""
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/settings/searchable-attributes" root-url meilisearch-index)
       :method :get
       :use-connection-pool nil
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun get-stats (root-url meilisearch-index meilisearch-api-key)
  "Returns a collections of stats from the `MEILISEARCH-INDEX' in the database."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/stats" root-url meilisearch-index)
       :method :get
       :headers `(("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :use-connection-pool nil
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun set-filter-attributes (filter-attributes root-url meilisearch-index meilisearch-api-key)
  "Sets the filter attributes in the Meilisearch database for the `MEILISEARCH-INDEX'.
`FILTER-ATTRIBUTES' should be a string consisting of what 'attributes' you want
the database to allow as filter when searching."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/settings" root-url meilisearch-index)
       :method :patch
       :headers `(("Content-Type" . "application/json")
                  ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :content (format nil "{ \"filterableAttributes\": [ \"~a\" ] }" filter-attributes)
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun set-ranking-rules (ranking-attributes root-url meilisearch-index meilisearch-api-key)
  "Sets the order in which the database ranks the search results.
`RANKING-ATTRIBUTES' should be a string consisting of what 'attributes' you want
the database to prioritise, and in what in order."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/settings" root-url meilisearch-index)
       :method :patch
       :use-connection-pool nil
       :headers `(("Content-Type" . "application/json")
                  ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :content (format nil "~a" ranking-attributes)
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun set-search-attributes (search-attributes root-url meilisearch-index meilisearch-api-key)
  "Sets the 'search attributes' in the Meilisearch database.
These attributes determine which values can be used when searching the database.
An example of what the `SEARCH-ATTRIBUTES' should look like when passed to this
function is as follows, '[ \"title\", \"genre\" ]' (replace single quotes with
double, used here due to formatting of doc-string)."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/settings/searchable-attributes" root-url meilisearch-index)
       :method :put
       :use-connection-pool nil
       :headers `(("Content-Type" . "application/json")
	          ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :content (format nil "~a" search-attributes)
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun submit-entry (data root-url meilisearch-index meilisearch-api-key)
  "Submits recipe to the Meilisearch database, use to create and edit entry.
`DATA' needs to be serialised to JSON before calling this function."
  (handler-case
      (dexador:request
       (format nil "~a/indexes/~a/documents" root-url meilisearch-index)
       :method :post
       :use-connection-pool nil
       :headers `(("Content-Type" . "application/json")
                  ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
       :content data
       :verbose nil)
    (error (e) (format nil "Search Instance Error: ~a" e))))

(defun submit-batch-entries (entries root-url meilisearch-index meilisearch-api-key)
  "Submits a batch of serialised (JSON) entries to the Meilisearch instance.
Make sure you serialise the list of`ENTRIES' to JSON before calling this function."
  (let ((success 0)
        (fail 0))
    (loop :for entry :in entries :do
      (handler-case
          (progn
            (dexador:request
             (format nil "~a/indexes/~a/documents" root-url meilisearch-index)
             :method :post
             :use-connection-pool t
             :keep-alive t
             :headers `(("Content-Type" . "application/json")
                        ("Authorization" . ,(format nil "Bearer ~a" meilisearch-api-key)))
             :content entry
             :verbose nil)
            (incf success))
        (error (e)
          (incf fail)
          (format *error-output* "Search Instance Error: ~a" e))))
    (format nil "Batch Upload Complete: Successful: ~a Failed: ~a" success fail)))
      
