(in-package #:cl-user)
(defpackage #:storage
  (:use #:cl
        #:larder.web.config
        #:app-constants
        #:models)
  (:local-nicknames (#:config #:larder.web.config))
  (:export #:clear-storage
           #:create-user-dir
           #:get-file-names
           #:init-storage
           #:make-path
           #:open-text-file
           #:open-binary-file
           #:store-file
           #:store-text-file
           #:to-base64-string
           #:read-from-base64-string))
(in-package #:storage)

#| Built-In Functions to Use Alongside this Package.
================================================================================
There are functions which already provide the necessary functionality. This
package doesn't aim to re-write them here. With that said, here is a list of
functions you might find helpful. You will most like use these alongside this
package in the various `DEFROUTE' macros defined in the `LARDER.WEB.WEB'
package.

(cl:probe-file)
(uiop:directory-files)
(file-namestring)
(cl-fad:delete-directory-and-files)
(rename-file)
(delete-file)
|#

;; Should move this to the utils package at some point.
(defun to-base64-string (object)
  (cl-base64:string-to-base64-string (prin1-to-string object)))

;; And this.
(defun read-from-base64-string(string)
  (values (read-from-string
	   (cl-base64:base64-string-to-string string))))

(defun clear-storage ()
  "Deletes the storage/ directory and everything in it."
  (cl-fad:delete-directory-and-files +storage-dir+))

(defun create-user-dir (account-id)
  "Creates the standard directory structure for a 'user' `ACCOUNT'
The directory structure is as follows,

*APPLICATION-ROOT*/storage/`ACCOUNT-ID'/
├── recipe
└── thumbnail

2 directories, 0 files"
  (ensure-directories-exist
   (merge-pathnames (format nil "~a/recipe/" account-id) +user-dir+))
  (ensure-directories-exist
   (merge-pathnames (format nil "~a/thumbnail/" account-id) +user-dir+))
  (uiop:copy-file (make-path "static/default-assets/profile.png")
                  (make-path (format nil "storage/user/~a/profile" account-id))))

(defun get-file-names (filenames)
  "Returns a list of filenames, parsed from a list of paths."
  (mapcar #'(lambda (x) (file-namestring x)) filenames))

(defun init-storage ()
  "Creates the standard directory structure used to store media/asset files.
This should only be run during the website's 'first run', to set up the system.
The directory structure is as follows,

*APPLICATION-ROOT*/storage/
├── media
├── page
├── recipe
├── thumbnail
└── users

5 directories, 0 files"
  (uiop:copy-file (make-path "static/default-assets/site-logo.png")
                  (make-path "static/images/site-logo"))
  (uiop:copy-file (make-path "static/default-assets/favicon.png")
                  (make-path "static/images/favicon.png"))
  (ensure-directories-exist +storage-dir+)
  (ensure-directories-exist +page-dir+)
  (ensure-directories-exist +media-dir+)
  (ensure-directories-exist +recipe-dir+)
  (ensure-directories-exist +thumb-dir+)
  (ensure-directories-exist +user-dir+)
  ;; Home is a built-in page and cannot be deleted. The Home page is copied here
  ;; to overwrite the default message, stating how to setup the website.
  (uiop:copy-file (make-path "static/default-assets/home.html")
                  (format nil "~a~a" +page-dir+ +home-page-uuid+)))

(defun make-path (path)
  "Merges `PATH' with `*APPLICATION-ROOT*' (use to reduce typing)."
  (merge-pathnames path config:*application-root*))

(defun open-binary-file (path)
  "Reads the file stored in at `PATH'."
  (with-open-file (stream path :element-type '(unsigned-byte 8))
    (let* ((length (file-length stream))
	   (buffer (make-array length :element-type '(unsigned-byte 8))))
      (read-sequence buffer stream)
      (values buffer length))))

(defun open-text-file (path)
  "Reads the text (usually either .html or .md) file at `PATH'."
   (with-open-file (stream path)
     (let ((data (make-string (file-length stream))))
       (read-sequence data stream)
       data)))

(defun store-text-file (filepath data)
  "Stores text `DATA' at `FILEPATH'."
  (with-open-file (stream
                   filepath
                   :direction :output
                   :if-does-not-exist :create
                   :if-exists :supersede)
    (format stream "~a" data)))

(defun store-file (path data)
  "Stores the uploaded file at `PATH'."
  (let ((filepath (ensure-directories-exist path)))
    (cond ((equal (type-of (car data)) 'SB-SYS:FD-STREAM)
	   (uiop:copy-file (slot-value (car data) 'pathname) filepath))
	  (t (with-open-file (stream
			      filepath
			      :direction :output
			      :if-does-not-exist :create
			      :element-type '(unsigned-byte 8)
			      :if-exists :supersede)
	       (write-sequence (slot-value (car data) 'vector) stream))))))
