(in-package #:cl-user)
(defpackage #:larder.web.web
  (:use #:cl
        #:caveman2
        #:larder.web.config
        #:larder.web.view
        #:larder.web.db
        #:hermetic
        #:diskspace
        #:zip
        #:utils
        #:validation
        #:storage
        #:pagination
        #:app-constants
        #:status-codes
        #:serial
        #:meilisearch)
  (:local-nicknames (#:db #:larder.web.db)
                    (#:jzon #:com.inuoe.jzon)
                    (#:search #:meilisearch))
  (:export #:*web*))
(in-package #:larder.web.web)

(syntax:use-syntax :annot) ; for @route annotation

;; APPLICATION
;; =============================================================================

(defclass <web> (<app>) ())
(defvar *web* (make-instance '<web>))
(clear-routing-rules *web*)

#| ROUTING RULES
=============================================================================
The routes are broken down into the following sections:
INITIAL SET-UP / FIRST RUN
LOG-IN MANAGEMENT
SIGN-UP MANAGEMENT
ACCOUNT BACK-END (I.E. logged in users/administrators/owner)
SITE SETTINGS
PAGE CRUD (Owner/Admin controlled 'pages', About, Contact, Home Etc.)
RECIPE CRUD (Individual account recipes, private by default)
OWNER/ADMIN RECIPE CRUD (Public facing recipes)
FAVOURITES (Starring/Favouriting recipes for all account types)
STORAGE FILES
SITE MAP
DANGER ZONE
ERROR PAGES
UNSORTED ROUTES
|#

(defroute ("/" :method :GET) ()
  (if (db:get-page :title +home-page-title+)
      ;; Site is set up and running as intended.
      (let ((page-data (db:get-page :title +home-page-title+)))
        (render "page.html"
                (append (if (hermetic:logged-in-p) (auth:auth-user-data))
                        `(:alert ,(utils:get-and-reset-alert)
                          :system-data ,(db:get-system-data)
                          :data
                          ,(storage:open-text-file
                            (make-path
                             (format nil "~a~a"
                                     +page-dir+ (models::page-page-id page-data))))))))
      ;; Site still needs to go through it's first-run/set-up.
      (render "index.html")))

;; INITIAL SET-UP / FIRST RUN
;; =============================================================================

(defroute ("/setup" :method :GET) () ; Used during the site's first-run.
  (if (db:site-settings-exist?)
      (throw-code +forbidden+) ; Becomes inaccessible after the site is set up.
      (render "setup.html"
              `(:token ,(auth:csrf-token)
                :alert ,(utils:get-and-reset-alert)
                :environment ,(larder.web.config:appenv)))))

(defroute ("/run-setup" :method :POST) () ; Used to set up site on firtst-run.
  (destructuring-bind
      (&key authenticity-token site-name username display-name password
         allow-sign-up show-site-logo search-url search-index search-api-key &allow-other-keys)
      (utils:request-params (lack.request:request-body-parameters ningle:*request*))
    (cond ((not (string= authenticity-token (auth:csrf-token)))
           (throw-code +forbidden+)) ; Inaccessible after the site is set up.
          ((find t (mapcar
		    #'validation:string-is-nil-or-empty?
		    `(,site-name ,username ,display-name ,password)))
           (utils:set-alert "Parts of form not completed." +err+)
           (redirect "/setup"))
          (t (db:init-db site-name username display-name password
                         (if allow-sign-up t nil)
                         (if show-site-logo t nil)
                         (if (validation:string-is-nil-or-empty? search-url)
                             nil
                             search-url)
                         (if (validation:string-is-nil-or-empty? search-index)
                             nil
                             search-index)
                         (if (validation:string-is-nil-or-empty? search-api-key)
                             nil
                             search-api-key))
             (storage:init-storage)
             (storage:create-user-dir
              (models::account-account-id (db:get-account :username username)))
             (db:create-page +home-page-title+
                             (models::account-account-id (db:get-account :username username))
                             +home-page-uuid+
                             "on" ; Show in nav. menu.
                             nil  ; Page cannot be deleted.
                             "on" ; Page is public
                             )
             ;; This is needed so the system can use 'sudo' commands with the
             ;; Postgresql database from the website -- and not a dev. SSH'd
             ;; into the server. The most notable task is creating dumps (part
             ;; of snapshot feature) of the database whilst logged into the
             ;; website (again, not SSH'd to the server).
             (storage:store-text-file
              "~/.pgpass" (format nil "localhost:5432:~a:~a:~a"
                                  (larder.web.config:database-name)
                                  (larder.web.config:database-username)
                                  (larder.web.config:database-password)))
             (utils:set-alert "All set-up. Log-in to continue." +suc+)
             (redirect "/login")))))

;; LOG-IN MANAGEMENT
;; =============================================================================

(defroute ("/login" :method :GET) ()
  (hermetic:auth (:user :administrator :owner)
		 ;; Authorised
		 (redirect "/dashboard")
		 ;; Not Authorised
		 (render "login.html"
			 `(:token ,(auth:csrf-token)
			   :system-data ,(db:get-system-data)
			   :alert ,(utils:get-and-reset-alert)))))

(defroute ("/attempt-login" :method :POST) ()
  (destructuring-bind
      (&key username password authenticity-token &allow-other-keys)
      (utils:request-params (lack.request:request-body-parameters ningle:*request*))
    (cond ((not (string= authenticity-token (auth:csrf-token)))
	   (throw-code +forbidden+))
          ((null (db:get-account :username username))
           (utils:set-alert "Account not found" +inval+)
           (redirect "/login"))
          ((not (eql t (models::account-active (db:get-account :username username))))
           (utils:set-alert "Account is currently deactivated." +inval+)
           (redirect "/login"))
          (t ;; The 'Failed log-in attempt' and 'No user found' sections are
             ;; kinda redundant at this point. They need to remain, though,
             ;; because they are part of the hermetic:login macro. If the 'is
             ;; active' stuff (above) is removed in a future commit, the account
             ;; checks above can be removed. The code can then fall back onto
             ;; the 'Failed log-in attempt' and 'No user found' sections
             ;; below. Yes, this is a bit messy but it is what it is...
           (let ((params (list :|username| username :|password| password)))
	       (hermetic:login
	        params
	        ;; Successful log-in attempt.
	        (progn
	          (setf
	           ;; Set session Id. to the logged in user. Use
	           (gethash :id ningle:*session*) (auth:get-account-db-id username)
	           ;; Set the users password (for session)
	           (gethash :password ningle:*session*) password)
	          (redirect "/dashboard"))
	        ;; Failed log-in attempt.
	        (progn (utils:set-alert "Incorrect details provided." +inval+)
		       (redirect "/login"))
	        ;; No user found.
	        (progn (utils:set-alert "Can't find that user." +err+)
		       (redirect "/login"))))))))

(defroute ("/logout" :method :POST) ()
  (destructuring-bind
      (&key authenticity-token &allow-other-keys)
      (utils:request-params (lack.request:request-body-parameters ningle:*request*))
    (if (not (string= authenticity-token (auth:csrf-token)))
        (throw-code +forbidden+)
        (hermetic::logout
         ;; Successful log-out.
         (progn (utils:flash-session-hash :id ningle:*session*)
                (utils:set-alert "You have logged out." +suc+)
                (redirect "/"))
         ;; Failed log-out
         (progn (utils:set-alert "Failed to log out." +err+)
                (redirect "/"))))))

;; SIGN-UP MANAGEMENT
;; =============================================================================

(defroute ("/signup" :method :GET) ()
  (let ((sys-setting (db:get-site-settings)))
    (if (models::site_settings-allow-sign-up sys-setting)
        (hermetic:auth
            (:user :administrator :owner)
            ;; Authorised
            (redirect "/dashboard")
            ;; Not Authoried
            (render "signup.html" `(:token ,(auth:csrf-token)
                                    :site-settings ,sys-setting
                                    :alert ,(utils:get-and-reset-alert))))
        (throw-code +not-found+))))

(defroute ("/attempt-signup" :method :POST) ()
  (destructuring-bind
      (&key authenticity-token username display-name password password-check
       &allow-other-keys)
      (utils:request-params (lack.request:request-body-parameters ningle:*request*))
    (cond ((not (string= authenticity-token (auth:csrf-token)))
	   (throw-code +forbidden+))
          ((find t (mapcar #'validation:string-is-nil-or-empty?
                           `(,username ,display-name ,password ,password-check)))
           (utils:set-alert "Missing required data." +miss+)
           (redirect "/signup"))
          ((not (string= password password-check))
           (utils:set-alert "Passwords don't match." +inval+)
           (redirect "/signup"))
          ((db:get-account :username username)
           (utils:set-alert "Username already taken." +warn+)
           (redirect "/signup"))
          (t (db:create-account username display-name password :user t)
             (storage:create-user-dir
              (models::account-account-id (db:get-account :username username)))
             (utils:set-alert "Account created." +suc+)
             (render "/signup-success.html" `(:token ,(auth:csrf-token)
		                              :system-data ,(db:get-system-data)
                                              :username ,username
                                              :display-name ,display-name
                                              :password ,password
		                              :alert ,(utils:get-and-reset-alert)))))))

;; ACCOUNT BACK-END
;; =============================================================================

(defroute ("/dashboard" :method :GET) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (let ((user (auth:get-current-user)))
        (render #P"dashboard.html"
		`(:alert ,(utils:get-and-reset-alert)
                  :pages ,(db:get-all-pages)
                  :media-files ,(db:get-all-files)
                  :site-recipes ,(db:get-all-recipes)
                  :disk-info ,(first (diskspace:list-all-disk-info t))
                  :token ,(auth:csrf-token)
                  :roles ,(auth:get-user-roles)
                  :user ,user
                  :favourites
                  ,(db:get-favourite-recipes (models::account-account-id user)
                                             :limit 5 :reverse t)
                  :user-recipes
                  ,(db:get-user-recipes
                    (models::account-account-id (auth:get-current-user)))
                  :system-data ,(db:get-system-data))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/user/edit" :method :GET) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (render #P"edit-account.html"
	      (append (auth:auth-user-data)
		      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/user/edit/update-profile" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token profile-image &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? (cadr profile-image))
               (utils:set-alert "No file uploaded." +inval+)
               (redirect "/user/edit"))
              ((not (str:contains? "image" (caddr profile-image) :ignore-case t))
               (utils:set-alert "Uploaded file is not an image." +inval+)
               (redirect "/user/edit"))
              (t (let ((filepath
                         (make-path (format nil "storage/user/~a/profile"
                                            (models::account-account-id
                                             (auth:get-current-user))))))
                   (storage:store-file filepath profile-image)
                   (utils:scale-image filepath 512 512)
                   (utils:set-alert "Profile image updated." +suc+)
                   (redirect "/user/edit")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/user/edit/display-name" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token display-name &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? display-name)
               (utils:set-alert "Missing new Display Name." +miss+)
               (redirect "/user/edit"))
              (t (db:update-account
                  (models::account-account-id (auth:get-current-user))
                  :display-name display-name)
                 (utils:set-alert "Display Name updated." +suc+)
                 (redirect "/user/edit"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/user/edit/password" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token new-password password-check &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((find t (mapcar #'validation:string-is-nil-or-empty?
                               `(,new-password ,password-check)))
               (utils:set-alert "Missing Password data." +miss+)
               (redirect "/user/edit"))
              ((not (string-equal new-password password-check))
               (utils:set-alert "Passwords don't match." +inval+)
               (redirect "/user/edit"))
              (t (db:update-account
                  (models::account-account-id (auth:get-current-user))
                  :new-password new-password)
                 (utils:set-alert "Password updated." +suc+)
                 (redirect "/user/edit"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/user/delete" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token delete-account-check &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? delete-account-check)
               (utils:set-alert "Missing Delete Account Check." +miss+)
               (redirect "/user/edit"))
              ((not (string-equal
                     delete-account-check
                     (models::account-username (auth:get-current-user))))
               (utils:set-alert "Username entered doesn't match account username." +miss+)
               (redirect "/user/edit"))
              (t (let ((acc-id-delete
                         (models::account-account-id (auth:get-current-user))))
                   (hermetic::logout
                    ;; Successful log-out.
                    (progn
                      (db:delete-account acc-id-delete)
                      (utils:set-alert "Account deleted." +suc+)
                      (utils:flash-session-hash :id ningle:*session*)
                      (redirect "/"))
                    ;; Failed log-out
                    (progn (utils:set-alert "Failed to log out." +err+)
                           (redirect "/")))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/accounts" :method :GET) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (render #P"accounts-index.html"
	      (append (auth:auth-user-data)
		      `(:alert ,(utils:get-and-reset-alert)
                        :accounts ,(db:get-all-accounts) ; paginate it?
                        :system-data ,(db:get-system-data))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/accounts/user-to-admin" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token user-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? user-id)
               (utils:set-alert "No account provided." +miss+)
               (redirect "/accounts"))
              ((not (db:get-account :account-id user-id))
               (utils:set-alert "Account not found." +inval+)
               (redirect "/accounts"))
              (t (db:update-account-type user-id :admin t)
                 (utils:set-alert "Updated user to admin." +suc+)
                 (redirect "/accounts"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/accounts/admin-to-user" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token admin-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? admin-id)
               (utils:set-alert "No account provided." +miss+)
               (redirect "/accounts"))
              ((not (db:get-account :account-id admin-id))
               (utils:set-alert "Account not found." +inval+)
               (redirect "/accounts"))
              (t (db:update-account-type admin-id :user t)
                 (utils:set-alert "Admin. account changed to User." +suc+)
                 (redirect "/accounts"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/accounts/set-account-status" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token account-id account-status &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? account-id)
               (utils:set-alert "No account Id. provided." +miss+)
               (redirect "/accounts"))
              ((not (or (string-equal "activate" account-status)
                        (string-equal "deactivate" account-status)))
               (utils:set-alert "Invalid Account Status used." +inval+)
               (redirect "/accounts"))
              ((not (db:get-account :account-id account-id))
               (utils:set-alert "Account not found." +inval+)
               (redirect "/accounts"))
              (t (db:set-account-status
                  account-id (if (string= "activate" account-status) t nil))
                 (utils:set-alert "Account updated." +suc+)
                 (redirect "/accounts"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/accounts/delete-account" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token account-id delete-account-check &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((find t
                     (mapcar #'validation:string-is-nil-or-empty?
                             `(,account-id ,delete-account-check)))
               (utils:set-alert "Missing Account data (Id. or Username)." +miss+)
               (redirect "/accounts"))
              ((not (string-equal delete-account-check
                                  (models::account-username
                                   (db:get-account :account-id account-id))))
               (utils:set-alert
                "Username entered doesn't match account username." +miss+)
               (redirect "/accounts"))
              (t (cl-fad:delete-directory-and-files
                  (storage:make-path (format nil "storage/user/~a/" account-id)))
                 (db:delete-account account-id)
                 (utils:set-alert "Account deleted." +suc+)
                 (redirect "/accounts"))))
      ;; Not Authorised
      (throw-code +forbidden+)))


;; SITE SETTINGS
;; =============================================================================

(defroute ("/site-settings" :method :GET) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (let* ((snippet-path
               (storage:make-path "storage/site-wide-snippet.html"))
             (snippet-data (if (cl:probe-file snippet-path)
                               (storage:open-text-file snippet-path)
                               nil)))
        (render "site-settings.html"
                (append (auth:auth-user-data)
                        `(:alert ,(utils:get-and-reset-alert)
                          :site-wide-snippet ,snippet-data
                          :pages ,(db:get-all-pages)
                          :settings ,(db:get-site-settings)
                          :system-data ,(db:get-system-data)))))
      ;; Not Authoried
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-site-name" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key site-name authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? site-name)
	       (utils:set-alert
                "No value provided. Site name not changed." +miss+)
	       (redirect "/site-settings"))
	      (t (db:update-site-name site-name)
		 (utils:set-alert "Site name updated." +suc+)
		 (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-site-logo" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key site-logo authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? (cadr site-logo))
	       (utils:set-alert "No file provided. Site logo not uploaded." +miss+)
               (redirect "/site-settings"))
              ((not (str:contains? "image" (caddr site-logo) :ignore-case t))
               (utils:set-alert "Uploaded file is not an image." +inval+)
               (redirect "/site-settings"))
	      (t (storage:store-file (make-path "static/images/site-logo") site-logo)
		 (utils:set-alert "Site Logo updated." +suc+)
		 (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-show-site-logo" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key show-site-logo authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              (t (db:update-show-site-logo-setting show-site-logo)
	         (utils:set-alert "Site Logo setting updated." +suc+)
	         (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-show-log-in" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key update-show-log-in authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              (t (db:update-show-log-in update-show-log-in)
	         (utils:set-alert "Log-In setting updated." +suc+)
	         (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-sign-up" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key allow-sign-up authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              (t (db:update-sign-up-setting allow-sign-up)
	         (utils:set-alert "Sign-Up setting updated." +suc+)
	         (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-search-integration" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
       (&key show-search authenticity-token &allow-other-keys)
       (utils:request-params (lack.request:request-body-parameters ningle:*request*))
       (cond ((not (string= authenticity-token (auth:csrf-token)))
              (throw-code +forbidden+))
             (t (db:update-show-search show-search)
                (utils:set-alert "Search integration setting updated." +suc+)
                (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-search-url" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key search-url authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
	      (t (if (validation:string-is-nil-or-empty? search-url)
                     (db:update-search-url nil)
                     (db:update-search-url search-url))
		 (utils:set-alert "Search URL updated." +suc+)
		 (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-search-index" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key search-index authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
	      (t (if (validation:string-is-nil-or-empty? search-index)
                     (db:update-search-index nil)
                     (db:update-search-index search-index))
		 (utils:set-alert "Search Index updated." +suc+)
		 (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-search-api-key" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key search-api-key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
	      (t (if (validation:string-is-nil-or-empty? search-api-key)
                     (db:update-search-api-key nil)
                     (db:update-search-api-key search-api-key))
		 (utils:set-alert "Search API Key update." +suc+)
		 (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-nav-menu" :method :POST) (&key _parsed)
  ;; The HTML form posting to this route has a hidden input element. This is so
  ;; the back-end triggers the process of looping through all the pages and
  ;; removing them from the Nav-Menu when NO PAGES are selected. Without this
  ;; hidden input, the back-end doesn't receive any Page data when none of them
  ;; are selected. So, if you uncheck all the pages, the system thinks nothing
  ;; has changed because it has received no data. Making sure at least one
  ;; (hidden) item is sent as part of the request is how the system works around
  ;; this constraint.
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              (t (db:update-nav-menu
                  (cdr (assoc "NAV-MENU" _parsed :test #'string=)))
	         (utils:set-alert "Nav. menu updated." +suc+)
	         (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-favicon" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key favicon-file authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
	(cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? (cadr favicon-file))
	       (utils:set-alert "No file provided. Favicon not uploaded." +miss+)
	       (redirect "/site-settings"))
              ((not (str:contains?
		     "image/png" (caddr favicon-file) :ignore-case t))
               (utils:set-alert "Uploaded file is not a PNG image." +inval+)
               (redirect "/site-settings"))
	      (t (let ((fav-path
                         (merge-pathnames #P"images/favicon.png" *static-directory*)))
                   (storage:store-file fav-path favicon-file)
		   (utils:set-alert "Favicon saved." +suc+)
		   (utils:run-bash-command
		    (format nil "convert ~a -resize 192x192\\> ~a" fav-path fav-path))
		   (redirect "/site-settings")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/site-settings/update-code-snippet" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key code-snippet authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              (t (storage:store-text-file
                  (merge-pathnames #P"site-wide-snippet.html" +storage-dir+)
                  code-snippet)
		 (utils:set-alert "Site-Wide snippet updated." +suc+)
		 (redirect "/site-settings"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

;; PAGE CRUD (Owner/Admin controlled 'pages', About, Contact, Home Etc.)
;; =============================================================================

(defroute ("/pages?:page?" :method :GET) (&key (page "1"))
   (let ((page-size 100))
     (flet ((render-page (pagination pages)
              (when (> (gethash :page pagination) (gethash :nb-pages pagination))
                (throw-code +not-found+))
              (render "pages.html"
                      (append (if (hermetic:logged-in-p) (auth:auth-user-data))
                              `(:alert ,(utils:get-and-reset-alert)
                                :system-data ,(db:get-system-data)
                                :pagination ,pagination
                                :root-url "pages"
                                :pages ,pages)))))
       (hermetic:auth
           (:administrator :owner)
           ;; Authorised
           (handler-case
               (let* ((pages (db:get-paginated-pages
                              (parse-integer page) page-size :title))
                      (pagination (pagination:make-pagination
                                   :page (read-from-string page)
                                   :page-size page-size
                                   :nb-elements (db:get-table-count 'models:page))))
                 (render-page pagination pages))
             (error (e)
               (format *error-output* "[ERROR] ~a" e)
               (throw-code +not-found+)))
           ;; Not Authorised
           (handler-case
               (let* ((pages (db:get-paginated-pages
                              (parse-integer page) page-size :title :only-public t))
                      (pagination (pagination:make-pagination
                                   :page (read-from-string page)
                                   :page-size page-size
                                   :nb-elements (db:get-pages-count :only-public t))))
                 (render-page pagination pages))
             (error (e)
               (format *error-output* "[ERROR] ~a" e)
               (throw-code +not-found+)))))))

(defroute ("/page/view/:id" :method :GET) (&key id)
  (when (or (validation:string-is-nil-or-empty? id)
            (not (db:get-page :page-id id)))
    (throw-code +not-found+))
  (flet ((render-page ()
           (let* ((page-data (db:get-page :page-id id))
                  (file-path
                    (make-path (format nil "~a~a" +page-dir+ (models::page-page-id page-data)))))
             (render
              "page.html"
              (append (if (hermetic:logged-in-p) (auth:auth-user-data))
                      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data)
                        :meta-data ,page-data
                        :data ,(storage:open-text-file file-path)))))))
    (hermetic:auth
        (:administrator :owner)
        ;; Authorised
        (render-page)
        ;; Not Authorised
        (if (eql t (models::page-public (db:get-page :page-id id)))
            (render-page)
            (throw-code +forbidden+)))))

(defroute ("/page/new" :method :GET) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (render "create-page.html"
              (append (auth:auth-user-data)
                      `(:alert ,(utils:get-and-reset-alert)
                        :sytem-data ,(db:get-system-data))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/page/create" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token title is-public in-nav-menu page-content &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (flet ((render-with-content ()
                 ;; Use to reduce duplicate code. Users can keep what they've
                 ;; written after an error has occurred.
                 (render "create-page.html"
                         (append (auth:auth-user-data)
                                 `(:alert ,(utils:get-and-reset-alert)
                                   :system-data ,(db:get-system-data)
                                   :data ,page-content)))))
          (cond ((not (string= authenticity-token (auth:csrf-token)))
                 (throw-code +forbidden+))
                ((validation:string-is-nil-or-empty? title)
                 (utils:set-alert "No title was entered." +miss+)
                 (render-with-content))
                ((db:get-page :title title)
                 (utils:set-alert "Page with that name already exists." +inval+)
                 (render-with-content))
                (t (let* ((account-id (models::account-account-id (auth:get-current-user)))
                          (pg-uuid (utils:lowercase-uuid title)))
                     (storage:store-text-file
                      (make-path (format nil "~a~a" +page-dir+ pg-uuid))
                      page-content)
                     (db:create-page
                      title account-id pg-uuid in-nav-menu t is-public)
		     (utils:set-alert "Page created. Great Success!" +suc+)
		     (redirect "/dashboard"))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/page/delete" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token page-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? page-id)
               (utils:set-alert "Page Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid? page-id))
               (utils:set-alert "Page Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-page :page-id page-id))
               (utils:set-alert "Could not find page in the database." +inval+)
               (redirect "/dashboard"))
              ((null (models::page-can-delete (db:get-page :page-id page-id)))
               (utils:set-alert "This page is not allowed to be deleted." +warn+)
               (redirect "/dashboard"))
              (t (delete-file (make-path (format nil "~a~a" +page-dir+ page-id)))
                 (db:delete-page :page-id page-id)
                 (utils:set-alert "Page deleted." +suc+)
                 (redirect "/dashboard"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/page/edit/:page-id" :method :GET) (&key page-id)
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (render "edit-page.html"
              (append (auth:auth-user-data)
                      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data)
                        :data ,(storage:open-text-file
                                (make-path (format nil "~a~a" +page-dir+ page-id)))
                        :meta-data ,(db:get-page :page-id page-id))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/page/edit-title" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token page-id title &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? page-id)
               (utils:set-alert "Page Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid?  page-id))
               (utils:set-alert "Page Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-page :page-id page-id))
               (utils:set-alert "Could not find page in the database." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? title)
               (utils:set-alert "Title is missing." +miss+)
               (redirect (format nil "/page/edit/~a" page-id)))
              (t (db:update-page page-id :title title)
                 (utils:set-alert "Page title updated." +suc+)
                 (redirect (format nil "/page/edit/~a" page-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/page/update-status" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token page-id is-public &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? page-id)
               (utils:set-alert "Page Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid?  page-id))
               (utils:set-alert "Invalid page Id." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-page :page-id page-id))
               (utils:set-alert "Unable to find page in the database." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? is-public)
               (utils:set-alert "Missing status data." +miss+)
               (redirect (format nil "/page/edit/~a" page-id)))
              (t (if (string= "private" is-public) ; Yes, is public so make private.
                     (db:update-page page-id :is-public t)
                     (db:update-page page-id :is-public nil))
		 (utils:set-alert "Page status update." +suc+)
		 (redirect (format nil "/page/edit/~a" page-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/page/update-content" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token page-id content &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? page-id)
               (utils:set-alert "Page Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid?  page-id))
               (utils:set-alert "Invalid page Id." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-page :page-id page-id))
               (utils:set-alert "Unable to find page in the database." +inval+)
               (redirect "/dashboard"))
              (t (storage:store-text-file
                  (make-path (format nil "~a~a" +page-dir+ page-id))
                  content)
		 (utils:set-alert "Page updated." +suc+)
		 (redirect (format nil "/page/edit/~a" page-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

;; RECIPE CRUD (Individual account recipes, private by default)
;; =============================================================================

(defroute ("/recipes?:page?" :method :GET) (&key (page "1"))
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (handler-case
          (let* ((user (auth:get-current-user))
                 (page-size 100)
                 (pagination (pagination:make-pagination
                              :page (read-from-string page)
                              :page-size page-size
                              :nb-elements
                              (length (db:get-user-recipes
                                       (models::account-account-id user))))))
            (when (> (gethash :page pagination) (gethash :nb-pages pagination))
              (throw-code +not-found+))
            (render #P"recipes-index.html"
		    `(:alert ,(utils:get-and-reset-alert)
                      :token ,(auth:csrf-token)
                      :roles ,(auth:get-user-roles)
                      :system-data ,(db:get-system-data)
                      :pagination ,pagination
                      :root-url "recipes"
                      :user ,user
                      :recipes
                      ,(db:get-paginated-user-recipes
                        (parse-integer page) page-size :title
                        (models::account-account-id user)))))
        (error (e)
          (format *error-output* "[ERROR] ~a" e)
          (throw-code +not-found+)))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/view/:recipe-id" :method :GET) (&key recipe-id)
  (when (or (validation:string-is-nil-or-empty? recipe-id)
            (not (validation:is-valid-uuid? recipe-id))
            (not (db:get-user-recipe recipe-id)))
    (throw-code +not-found+))
  (flet ((render-recipe ()
           (let* ((recipe-data (db:get-user-recipe recipe-id))
                  (account-id (models::user_recipe-account-id recipe-data))
                  (file-path (make-path
                              (format nil "~a~a/recipe/~a" +user-dir+ account-id recipe-id))))
             (if (cl:probe-file file-path)
                 (render "user-recipe.html"
                         (append (if (hermetic:logged-in-p)
		                     (auth:auth-user-data))
		                 `(:alert ,(utils:get-and-reset-alert)
                                   :system-data ,(db:get-system-data)
                                   :data ,(storage:open-text-file file-path))))
                 (throw-code +not-found+)))))
    (hermetic:auth
        (:user :administrator :owner)
        ;; Authorised
        (render-recipe)
        ;; Not Authorised
        (if (equal t (models::user_recipe-public (db:get-user-recipe recipe-id)))
            (render-recipe)
             (throw-code +forbidden+)))))


(defroute ("/recipe/new" :method :GET) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (render #P"create-user-recipe.html"
              (append (auth:auth-user-data)
                      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/create" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token title is-public recipe-content &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (flet ((render-recipe ()
                 (render #P"create-user-recipe.html"
                         (append (auth:auth-user-data)
                                 `(:alert ,(utils:get-and-reset-alert)
                                   :system-data ,(db:get-system-data)
                                   :data ,recipe-content)))))
          (cond ((not (string= authenticity-token (auth:csrf-token)))
                 (throw-code +forbidden+))
                ((validation:string-is-nil-or-empty? title)
                 (utils:set-alert "No title was entered." +miss+)
                 (render-recipe))
                ((> (length recipe-content) +user-recipe-max-upload-size+)
                 (utils:set-alert
                  (format nil "File above max. upload size of ~a Megabits (Mb)."
                        (* +user-recipe-max-upload-size+ 0.000008))
                  +inval+)
                 (render-recipe))
                (t (let ((account-id (models::account-account-id (auth:get-current-user)))
                         (user-recipe-id (utils:lowercase-uuid title)))
                     (db:create-user-recipe account-id title is-public user-recipe-id)
                     (storage:store-text-file
                      (make-path
                       (format nil "~a/~a/recipe/~a" +user-dir+ account-id user-recipe-id))
                      recipe-content)
		     (utils:set-alert "Recipe created. Great Success!" +suc+)
		     (redirect "/dashboard"))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/edit/:user-recipe-id" :method :GET) (&key user-recipe-id)
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (cond ((or (not (validation:is-valid-uuid? user-recipe-id))
                 (not (db:get-user-recipe user-recipe-id)))
             (throw-code +not-found+))
            (t (render #P"edit-user-recipe.html"
                       (append (auth:auth-user-data)
                               `(:alert ,(utils:get-and-reset-alert)
                                 :system-data ,(db:get-system-data)
                                 :data ,(storage:open-text-file
                                         (make-path
                                          (format nil "~a/~a/recipe/~a"
                                                  +user-dir+
                                                  (models::account-account-id
                                                   (auth:get-current-user))
                                                  user-recipe-id)))
                                 :meta-data ,(db:get-user-recipe user-recipe-id))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/edit-title" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token recipe-id title &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (db:get-user-recipe recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? title)
               (utils:set-alert "No title was entered." +miss+)
               (redirect (format nil "/recipe/edit/~a" recipe-id)))
              (t (db:update-user-recipe recipe-id :title title)
	         (utils:set-alert "Recipe title updated." +suc+)
	         (redirect (format nil "/recipe/edit/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/update-status" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token recipe-id is-public &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (db:get-user-recipe recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? is-public)
               (utils:set-alert "Missing status data." +miss+)
               (redirect (format nil "/recipe/edit/~a" recipe-id)))
              (t (if (string= "yes" is-public) ; Yes, is public, make private.
                     (db:update-user-recipe recipe-id :is-public nil)
                     (db:update-user-recipe recipe-id :is-public t))
		 (utils:set-alert "Recipe status updated." +suc+)
		 (redirect (format nil "/recipe/edit/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/update-content" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token recipe-id content &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (db:get-user-recipe recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              ((> (length content) +user-recipe-max-upload-size+)
               (utils:set-alert
                (format nil "File above max. upload size of ~a Megabits (Mb)."
                        (* +user-recipe-max-upload-size+ 0.000008))
                +inval+)
               (render #P"edit-user-recipe.html"
                       (append (auth:auth-user-data)
                               `(:alert ,(utils:get-and-reset-alert)
                                 :system-data ,(db:get-system-data)
                                 :meta-data ,(db:get-user-recipe recipe-id)
                                 :data ,content))))
              (t (storage:store-text-file
                  (make-path
                   (format nil "~a/~a/recipe/~a"
                           +user-dir+
                           (models::account-account-id (auth:get-current-user))
                           recipe-id))
                  content)
		 (utils:set-alert "Recipe updated." +suc+)
		 (redirect (format nil "/recipe/edit/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/recipe/delete" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token recipe-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (db:get-user-recipe recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              (t (utils:set-alert "Recipe deleted." +suc+)
                 (handler-case
                     (progn
                       (delete-file
                        (make-path
                         (format nil "~a/~a/recipe/~a"
                                 +user-dir+
                                 (models::account-account-id (auth:get-current-user))
                                 recipe-id)))
                       (db:delete-user-recipe recipe-id))
                   (error (e)
                     (format *error-output* "[ERROR] ~a" e)
                     (utils:set-alert
                      "Recipe deleted with errors (database or file name incorrect.)" +warn+)))
		 (redirect "/dashboard"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/customise/recipe/:recipe-id" :method :GET) (&key recipe-id)
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (cond ((null (db:get-recipe :recipe-id recipe-id))
             (throw-code +not-found+))
            (t (render "customise-recipe.html"
                       (append (auth:auth-user-data)
                               `(:alert ,(utils:get-and-reset-alert)
                                 :data ,(storage:open-text-file
                                         (storage:make-path
                                          (format nil "~a/~a" +recipe-dir+ recipe-id)))
                                 :meta-data ,(db:get-recipe :recipe-id recipe-id)
                                 :system-data ,(db:get-system-data))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

;; OWNER/ADMIN RECIPE CRUD (Public facing recipes in Meilisearch database)
;; =============================================================================

(defroute ("/view/recipe/:recipe-id" :method :GET) (&key recipe-id)
  (cond ((or (validation:string-is-nil-or-empty? recipe-id)
             (not (validation:is-valid-uuid? recipe-id))
             (not (db:get-recipe :recipe-id recipe-id)))
         (throw-code +not-found+))
        (t (flet ((render-with-content (user-data)
                    (render "site-recipe.html"
                            (append (when user-data user-data)
		                    `(:alert ,(utils:get-and-reset-alert)
                                      :system-data ,(db:get-system-data)
                                      :meta-data ,(db:get-recipe :recipe-id recipe-id)
                                      :data
                                      ,(storage:open-text-file
                                        (make-path
                                         (format nil "~a~a" +recipe-dir+ recipe-id))))))))
             (hermetic:auth
                 (:administrator :owner)
                 ;; Authorised
                 (let* ((user (auth:get-current-user))
                        (roles (auth:get-user-roles))
                        (token (auth:csrf-token))
                        (is-favourite
                          (db:is-favourite? (models::account-account-id user) recipe-id))
                        (auth-user-data (list :token token
                                              :roles roles
                                              :user user
                                              :is-favourite is-favourite)))
                   (render-with-content auth-user-data))
                 ;; Not Authorised
                 (if (models::recipe-public (db:get-recipe :recipe-id recipe-id))
                     (render-with-content nil)
                     (throw-code +forbidden+)))))))

(defroute ("/new/recipe" :method :GET) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (render "create-recipe.html"
              (append (auth:auth-user-data)
                      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/create/recipe" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token title source ingredients page-content thumbnail
           &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (flet ((render-with-content ()
                 ;; Use to reduce duplicate code. Users can keep what they've
                 ;; written after an error has occurred.
                 (render "create-recipe.html"
                         (append (auth:auth-user-data)
                                 `(:alert ,(utils:get-and-reset-alert)
                                   :system-data ,(db:get-system-data)
                                   :title ,title
                                   :source ,source
                                   :ingredients ,ingredients
                                   :data ,page-content)))))
          (cond ((not (string= authenticity-token (auth:csrf-token)))
                 (throw-code +forbidden+))
                ((validation:string-is-nil-or-empty? title)
                 (utils:set-alert "No title was entered." +miss+)
                 (render-with-content))
                ((and (not (validation:string-is-nil-or-empty? (cadr thumbnail)))
                      (not (str:contains? "image" (caddr thumbnail))))
                 (utils:set-alert "Thumbnail uploaded is not an image." +inval+)
                 (render-with-content))
                (t (let* ((account-id (models::account-account-id (auth:get-current-user)))
                          (site-settings (db:get-site-settings))
                          (recipe-uuid (utils:lowercase-uuid title))
                          (page-path (make-path (format nil "~a~a" +recipe-dir+ recipe-uuid)))
                          (thumbnail-uuid
                            (if (validation:string-is-nil-or-empty? (cadr thumbnail))
                                nil
                                (utils:lowercase-uuid (cadr thumbnail)))))
                     (when thumbnail-uuid
                       (let* ((thumb-path
                                (make-path (format nil "~a~a" +thumb-dir+ thumbnail-uuid)))
                              (command
                                (format nil "convert ~a -resize 512x512\\> ~a"
                                        thumb-path thumb-path)))
                         (storage:store-file thumb-path thumbnail)
                         (utils:run-bash-command command)))
                     (storage:store-text-file page-path page-content)
                     (db:create-recipe
                      account-id
                      title
                      recipe-uuid
                      (if (validation:string-is-nil-or-empty? source) nil source)
                      (if (validation:string-is-nil-or-empty? ingredients) nil ingredients)
                      thumbnail-uuid)
                     (utils:set-alert "Recipe created. Great Success!" +suc+)
                     (when (and (models::site_settings-search-url site-settings)
                                (models::site_settings-search-index site-settings)
                                (models::site_settings-search-api-key site-settings))
                       (let ((result
                               (search:submit-entry
                                (serial:recipe-to-json
                                 recipe-uuid title thumbnail-uuid ingredients
                                 (quri:uri-domain (quri:uri source))
                                 (if (validation:string-is-nil-or-empty? source) nil source))
                                (models::site_settings-search-url site-settings)
                                (models::site_settings-search-index site-settings)
                                (models::site_settings-search-api-key site-settings))))
                         (when (str:contains? "error" result) (utils:set-alert result +err+))))
		     (redirect "/dashboard"))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/delete/recipe" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token recipe-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid? recipe-id))
               (utils:set-alert "Recipe Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-recipe :recipe-id recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              (t (let* ((recipe-to-delete (db:get-recipe :recipe-id recipe-id))
                        (thumb-id (models::recipe-thumbnail-id recipe-to-delete))
                        (recipe-path (make-path (format nil  "~a~a" +recipe-dir+ recipe-id)))
                        (thumb-path (make-path (format nil "~a~a" +thumb-dir+ thumb-id)))
                        (site-settings (db:get-site-settings)))
                   (if (cl:probe-file recipe-path) (delete-file recipe-path))
                   (when thumb-id
                     (if (cl:probe-file thumb-path) (delete-file thumb-path)))
                   (db:delete-recipe recipe-id)
                   (when (and (models::site_settings-search-url site-settings)
                              (models::site_settings-search-index site-settings)
                              (models::site_settings-search-api-key site-settings))
                     (let ((result
                             (search:delete-entry
                              recipe-id
                              (models::site_settings-search-url site-settings)
                              (models::site_settings-search-index site-settings)
                              (models::site_settings-search-api-key site-settings))))
                       (when (str:contains? "error" result) (utils:set-alert result +err+))))
                   (utils:set-alert "Recipe deleted." +suc+)
                   (redirect "/dashboard")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/edit/recipe/:recipe-id" :method :GET) (&key recipe-id)
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (cond ((or (not (validation:is-valid-uuid? recipe-id))
                 (not (db:get-recipe :recipe-id recipe-id)))
             (throw-code +not-found+))
            (t (render "edit-recipe.html"
                       (append (auth:auth-user-data)
                               `(:alert ,(utils:get-and-reset-alert)
                                 :system-data ,(db:get-system-data)
                                 :data ,(storage:open-text-file
                                         (format nil "~a~a" +recipe-dir+ recipe-id))
                                 :meta-data ,(db:get-recipe :recipe-id recipe-id))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/edit/recipe-title" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token recipe-id title &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid? recipe-id))
               (utils:set-alert "Recipe Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-recipe :recipe-id recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? title)
               (utils:set-alert "No title was entered." +miss+)
               (redirect (format nil "/edit/recipe/~a" recipe-id)))
              (t (db:update-recipe recipe-id :title title)
                 (utils:set-alert "Recipe title updated." +suc+)
                 (let ((site-settings (db:get-site-settings)))
                   (when (and (models::site_settings-search-url site-settings)
                              (models::site_settings-search-index site-settings)
                              (models::site_settings-search-api-key site-settings))
                     (let ((recipe-to-update (db:get-recipe :recipe-id recipe-id)))
                       (setf (models::recipe-title recipe-to-update) title)
                       (let ((result (search:submit-entry
                                      (serial:recipe-obj-to-json recipe-to-update)
                                      (models::site_settings-search-url site-settings)       
                                      (models::site_settings-search-index site-settings)     
                                      (models::site_settings-search-api-key site-settings))))
                         (when (str:contains? "error" result) (utils:set-alert result +err+)))))
	           (redirect (format nil "/edit/recipe/~a" recipe-id))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/edit/recipe-content" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token recipe-id content &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. missing unable to update." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid? recipe-id))
               (utils:set-alert "Recipe Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-recipe :recipe-id recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              (t (storage:store-text-file (format nil "~a~a" +recipe-dir+ recipe-id) content)
	         (utils:set-alert "Recipe updated." +suc+)
	         (redirect (format nil "/edit/recipe/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/edit/recipe-status" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token recipe-id is-public &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? recipe-id)
               (utils:set-alert "Recipe Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid? recipe-id))
               (utils:set-alert "Recipe Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-recipe :recipe-id recipe-id))
               (utils:set-alert "Unable to find recipe in the database." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? is-public)
               (utils:set-alert "Recipe status is missing." +miss+)
               (redirect (format nil "/edit/recipe/~a" recipe-id)))
              (t (utils:set-alert "Recipe status updated." +suc+)
                 (let ((site-settings (db:get-site-settings)))
                   (if (string= "yes" is-public)
                       (progn ; Yes, recipe is public, make private.
                           (db:update-recipe recipe-id :is-public nil)
                          (when (and (models::site_settings-search-url site-settings)
                                     (models::site_settings-search-index site-settings)
                                     (models::site_settings-search-api-key site-settings))
                            (let ((result
                                    (search:delete-entry
                                     recipe-id
                                     (models::site_settings-search-url site-settings)
                                     (models::site_settings-search-index site-settings)
                                     (models::site_settings-search-api-key site-settings))))
                              (when (str:contains? "error" result)
                                (utils:set-alert result +err+)))))
                       (progn ; Recipe is private, change to make it public.
                           (db:update-recipe recipe-id :is-public t)
                          (when (and (models::site_settings-search-url site-settings)
                                     (models::site_settings-search-index site-settings)
                                     (models::site_settings-search-api-key site-settings))
                            (let ((recipe-to-add (db:get-recipe :recipe-id recipe-id)))
                              (let ((result
                                      (search:submit-entry
                                       (serial:recipe-obj-to-json recipe-to-add)
                                       (models::site_settings-search-url site-settings)       
                                       (models::site_settings-search-index site-settings)     
                                       (models::site_settings-search-api-key site-settings))))
                                (when (str:contains? "error" result)
                                  (utils:set-alert result +err+))))))))
                 (redirect (format nil "/edit/recipe/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/edit/recipe-thumbnail" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token recipe-id thumbnail &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              ((or (not (validation:is-valid-uuid? recipe-id))
                   (not (db:get-recipe :recipe-id recipe-id)))
               (throw-code +not-found+))
              ((not (cl:probe-file
                     (format nil "~a~a"
                             +thumb-dir+
                             (models::recipe-thumbnail-id (db:get-recipe :recipe-id recipe-id)))))
               (utils:set-alert "Cannot find thumbnail with specified recipe Id." +inval+)
               (redirect (format nil "/edit/recipe/~a" recipe-id)))
              ((and (not (validation:string-is-nil-or-empty? (cadr thumbnail)))
                    (not (str:contains? "image" (caddr thumbnail))))
               (utils:set-alert "Thumbnail uploaded is not an image." +inval+)
               (redirect (format nil "/edit/recipe/~a" recipe-id)))
              (t (storage:store-file
                  (make-path (format nil "~a~a"
                                     +thumb-dir+
                                     (models::recipe-thumbnail-id
                                      (db:get-recipe :recipe-id recipe-id))))
                  thumbnail)
                 (utils:set-alert "Thumbnail updated." +suc+)
                 (redirect (format nil "/edit/recipe/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/update/recipes/set-status-with-source" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token status source-url &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              (t (let ((site-settings (db:get-site-settings))
                       (succeeded 0)
                       (failed 0))
                   ;; Search database code.
                   (when (and (models::site_settings-search-url site-settings)
                              (models::site_settings-search-index site-settings)
                              (models::site_settings-search-api-key site-settings))
                     (loop :for entry
                             :in (db:get-recipes-with-source
                                  (if (string= "null" source-url) nil source-url))
                           :do (if (string= "private" status)
                                   ;; Private (Remove from search database).
                                   (let ((result                                      
                                           (meilisearch:delete-entry
                                            (models::recipe-recipe-id entry)
                                            (models::site_settings-search-url site-settings)
                                            (models::site_settings-search-index site-settings)
                                            (models::site_settings-search-api-key
                                             site-settings))))
                                     (if (str:contains? "error" result)
                                         (incf failed)
                                         (incf succeeded)))
                                   ;; Public (Add to search database).
                                   (let ((result
                                           (meilisearch:submit-entry
                                            (serial:recipe-obj-to-json entry)
                                            (models::site_settings-search-url site-settings)
                                            (models::site_settings-search-index site-settings)
                                            (models::site_settings-search-api-key
                                             site-settings))))
                                     (if (str:contains? "error" result)
                                         (incf failed)
                                         (incf succeeded))))))
                   ;;  Main database code.
                   (if (string= "private" status)
                       ;; Private (make main database entries private).
                       (progn
                         (db:make-recipes-with-source-private
                          (if (string= "null" source-url) nil source-url))
                         (utils:set-alert
                          (format nil "Recipes set to private.
 ~a recipes deleted from search database, ~a failures." succeeded failed) +suc+))
                       ;; Public (make main Database entries public).
                       (progn
                         (db:make-recipes-with-source-public
                          (if (string= "null" source-url) nil source-url))
                         (utils:set-alert
                          (format nil "Recipes set to public.
 ~a recipes added to search database, ~a failures." succeeded failed)
                          +suc+)))
                   (redirect "/dashboard")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/update/recipes/make-private" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              (t (let ((site-settings (db:get-site-settings)))
                   (db:make-all-recipes-private)
                   (when (and (models::site_settings-search-url site-settings)
                              (models::site_settings-search-index site-settings)
                              (models::site_settings-search-api-key site-settings))
                     (meilisearch:delete-all-entries
                      (models::site_settings-search-url site-settings)
                      (models::site_settings-search-index site-settings)
                      (models::site_settings-search-api-key site-settings)))
                   (utils:set-alert "All recipes set to private and removed from search database."
                                    +suc+)
                   (redirect "/dashboard")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/update/recipes/make-public" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              (t (let ((site-settings (db:get-site-settings))
                       (succeeded 0)
                       (failed 0))
                   (db:make-all-recipes-public)
                   (when (and (models::site_settings-search-url site-settings)
                              (models::site_settings-search-index site-settings)
                              (models::site_settings-search-api-key site-settings))
                     (loop :for entry :in (db:get-all-recipes) :do
                       (let ((result
                               (meilisearch:submit-entry
                                (serial:recipe-obj-to-json entry)
                                (models::site_settings-search-url site-settings)
                                (models::site_settings-search-index site-settings)
                                (models::site_settings-search-api-key site-settings))))
                         (if (str:contains? "error" result) (incf failed)) (incf succeeded))))
                   (utils:set-alert
                    (format nil
                            "Recipes set to public in main database.
~a recipes added to search database ~a failures." succeeded failed) +suc+)
                   (redirect "/dashboard")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

;; FAVOURITES
;; =============================================================================

(defroute ("/favourites?:page?" :method :GET) (&key (page "1"))
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (handler-case
          (let* ((user (auth:get-current-user))
                 (page-size 100)
                 (pagination (pagination:make-pagination
                              :page (read-from-string page)
                              :page-size page-size
                              :nb-elements (db:get-table-count 'models:page))))
            (when (> (gethash :page pagination) (gethash :nb-pages pagination))
              (throw-code +forbidden+))
            (render #P"favourite-index.html"
		    `(:alert ,(utils:get-and-reset-alert)
                      :token ,(auth:csrf-token)
                      :roles ,(auth:get-user-roles)
                      :system-data ,(db:get-system-data)
                      :user ,user
                      :pagination ,pagination
                      :root-url "favourites"
                      :favourites
                      ,(db:get-paginated-favourites
                        (parse-integer page) page-size :recipe.recipe-id
                        (models::account-account-id user)))))
        (error (e)
          (format *error-output* "[ERROR] ~a" e)
          (throw-code +not-found+)))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/favourite/create" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token recipe-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((not (validation:is-valid-uuid? recipe-id))
               (utils:set-alert "The recipe Id. provided is invalid." +err+)
               (redirect "/dashboard"))
              ((not (db:get-recipe :recipe-id recipe-id))
               (utils:set-alert "The recipe could not be found in the database." +err+)
               (redirect "/dashboard"))
              ((not (cl:probe-file (format nil "~a~a" +recipe-dir+ recipe-id)))
               (utils:set-alert "Could not find the recipe's file." +err+)
               (redirect "/dashboard"))
              (t (db:create-favourite
                  (models::account-account-id (auth:get-current-user)) recipe-id)
                 (utils:set-alert "Recipe added to favourites." +suc+)
                 (redirect (format nil "/view/recipe/~a" recipe-id)))))
      ;; Not Authoised
      (throw-code +forbidden+)))

(defroute ("/favourite/delete" :method :POST) ()
  (hermetic:auth
      (:user :administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token recipe-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((not (validation:is-valid-uuid? recipe-id))
               (utils:set-alert "The recipe Id. provided is invalid." +err+)
               (redirect "/dashboard"))
              ((not (db:get-recipe :recipe-id recipe-id))
               (utils:set-alert "The recipe could not be found in the database." +err+)
               (redirect "/dashboard"))
              ((not (cl:probe-file (format nil "~a~a" +recipe-dir+ recipe-id)))
               (utils:set-alert "Could not find the recipe's file." +err+)
               (redirect "/dashboard"))
              (t (db:delete-favourite
                  (models::account-account-id (auth:get-current-user)) recipe-id)
                 (utils:set-alert "Recipe removed from favourites." +suc+)
                 (redirect (format nil "/view/recipe/~a" recipe-id)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

    

;; STORAGE FILES
;; =============================================================================

(defroute ("/storage?:page?" :method :GET) (&key (page "1"))
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (handler-case
          (let* ((user (auth:get-current-user))
                 (page-size 100)
                 (pagination (pagination:make-pagination
                              :page (read-from-string page)
                              :page-size page-size
                              :nb-elements (db:get-table-count 'models:file))))
            (when (> (gethash :page pagination) (gethash :nb-pages pagination))
              (throw-code +not-found+))
            (render #P"storage-index.html"
	            `(:alert ,(utils:get-and-reset-alert)
                      :token ,(auth:csrf-token)
                      :roles ,(auth:get-user-roles)
                      :system-data ,(db:get-system-data)
                      :user ,user
                      :pagination ,pagination
                      :root-url "storage"
                      :media-files ,(db:get-paginated-files
                                     (read-from-string page) page-size :title))))
        (error (e)
          (format *error-output* "[ERROR] ~a" e)
          (throw-code +not-found+)))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/storage/view/:file-id" :method :GET) (&key file-id)
  (when (or (not (validation:is-valid-uuid? file-id))
            (not (db:get-file file-id))
            (not (cl:probe-file (format nil "~a~a" +media-dir+ file-id))))
    (throw-code +not-found+))
  (flet ((return-file ()
           (let ((file (storage:open-binary-file (format nil "~a~a" +media-dir+ file-id)))
                 (mime (models::file-mime (db:get-file file-id))))
             `(200 (:content-type ,mime) ,file))))
    (hermetic:auth
        (:administrator :owner)
        ;; Authorised
        (return-file)
        ;; Not Authorised
        (if (not (models::file-public (db:get-file file-id)))
            (throw-code +forbidden+)
            (return-file)))))

(defroute ("/storage/thumbnail/:file-id" :method :GET) (&key file-id)
  (when (or (not (validation:is-valid-uuid? file-id))
            (not (db:get-file file-id))
            (not (cl:probe-file
                  (format nil "~a~a" +thumb-dir+
                          (models::file-thumbnail-id (db:get-file file-id))))))
    (throw-code +not-found+))
  (flet ((return-thumbnail ()
           (let* ((thumb-id (models::file-thumbnail-id (db:get-file file-id)))
                  (file (storage:open-binary-file
                         (format nil "~a~a" +thumb-dir+ thumb-id)))
                  (mime (models::file-mime (db:get-file file-id))))
             `(200 (:content-type ,mime) ,file))))
    (hermetic:auth
        (:administrator :owner)
        ;; Authorised
        (return-thumbnail)
        ;; Not Authorised
        (if (not (models::file-public (db:get-file file-id)))
            (throw-code +forbidden+)
            (return-thumbnail)))))

(defroute ("/storage/recipe-thumbnail/:thumbnail-id" :method :GET) (&key thumbnail-id)
  (when (or (not (validation:is-valid-uuid? thumbnail-id))
            (not (cl:probe-file (make-path (format nil "~a~a" +thumb-dir+ thumbnail-id)))))
    (throw-code +not-found+))
  (flet ((return-thumbnail ()
           ;; 'application/octet-stream' used as a catch-all, MIME not in
           ;; database and file extension not used when writing file to disk.
           `(200 (:content-type "application/octet-stream")
                 ,(storage:open-binary-file
                   (make-path (format nil "~a~a" +thumb-dir+ thumbnail-id))))))
    (hermetic:auth
        (:administrator :owner)
        ;; Authorised
        (return-thumbnail)
        ;; Not Authorised
        (if (not (models::recipe-public (db:get-recipe :thumbnail-id thumbnail-id)))
            (throw-code +forbidden+)
            (return-thumbnail)))))

(defroute ("/storage/multi-upload" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (let ((files (utils:separate-files-in-web-request
		      (lack.request:request-body-parameters ningle:*request*)
		      "STORAGE-FILES")))
          (cond ((not (string= authenticity-token (auth:csrf-token)))
                 (throw-code +forbidden+))
                ((validation:string-is-nil-or-empty? (caddr (car files)))
                 (utils:set-alert "No files uploaded" +miss+)
                 (redirect "/dashboard"))
                (t (loop :for item :in files :do
                  (let*
                      ((file-id (utils:lowercase-uuid (caddr item)))
                       (path (format nil "~a/~a" +media-dir+ file-id))
                       (title (caddr item))
                       (mime (cadddr item))
                       (account-id (models::account-account-id (auth:get-current-user)))
                       (thumbnail-id nil))
                    (storage:store-file path (cdr item))
                    (when (and (str:contains? "image" mime)
                               (not (str:contains? "svg" mime)))
                      (setf thumbnail-id (utils:lowercase-uuid (str:concat file-id "thumb")))
                      (utils:run-bash-command
                       (format nil "convert ~a -resize 512x512\\> ~a"
                               path (format nil "~a/~a" +thumb-dir+ thumbnail-id))))
                    (db:create-file title mime account-id thumbnail-id t file-id)))
                   (utils:set-alert "File upload complete." +suc+)
                   (redirect "/dashboard")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/storage/delete" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token file-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? file-id)
               (utils:set-alert "File Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid?  file-id))
               (utils:set-alert "File UUID is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-file file-id))
               (utils:set-alert "Cannot find file in the database." +err+)
               (redirect "/dashboard"))
              (t (let* ((file-to-delete (db:get-file file-id))
                        (file-id (models::file-file-id file-to-delete))
                        (thumb-id (models::file-thumbnail-id file-to-delete))
                        (file-path (format nil "~a~a" +media-dir+ file-id)))
                   (when thumb-id ; Thumbnails don't have own entry in database.
                     (let ((thumb-path (format nil "~a~a" +thumb-dir+ thumb-id )))
                       (if (cl:probe-file thumb-path) (delete-file thumb-path))))
                   (if (cl:probe-file file-path) (db:delete-file-entry file-id))
                   (delete-file file-path)
                   (utils:set-alert "File deleted." +suc+)
                   (redirect "/dashboard")))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/storage/edit-title" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token file-id title &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? file-id)
               (utils:set-alert "File Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? title)
               (utils:set-alert "Title Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid?  file-id))
               (utils:set-alert "File UUID is invalid." +inval+)
               (redirect "/dashboard"))
              ((not (db:get-file file-id))
               (utils:set-alert "Cannot find file in the database." +err+)
               (redirect "/dashboard"))
              (t (db:update-file file-id :title title)
                 (utils:set-alert "File updated." +suc+)
                 (redirect "/dashboard"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/storage/update-status" :method :POST) ()
  (hermetic:auth
      (:administrator :owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token file-id is-public &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation:string-is-nil-or-empty? file-id)
               (utils:set-alert "File Id. is missing." +miss+)
               (redirect "/dashboard"))
              ((not (validation:is-valid-uuid?  file-id))
               (utils:set-alert "File Id. is invalid." +inval+)
               (redirect "/dashboard"))
              ((validation:string-is-nil-or-empty? is-public)
               (utils:set-alert "Missing status data." +miss+)
               (redirect "dashboard"))
              ((not (db:get-file file-id))
               (utils:set-alert "Unable to find file in the database." +inval+)
               (redirect "/dashboard"))
              (t (if (string= "yes" is-public) ; Yes, file is public so make private.
                     (db:update-file file-id :is-public t)
                     (db:update-file file-id :is-public nil))
                 (utils:set-alert "File status updated." +suc+)
                 (redirect "/dashboard"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

;; SEARCH SECTION
;; =============================================================================

(defroute ("/search" :method :GET) ()
  (if (models::site_settings-show-search (db:get-site-settings))
    (render "search.html"
            (append (if (hermetic:logged-in-p) (auth:auth-user-data))
                    `(:alert ,(utils:get-and-reset-alert)
                      :system-data ,(db:get-system-data))))
    (throw-code +not-found+)))

;; SITE MAP
;; =============================================================================

(defroute ("/sitemap.xml" :method :GET) ()
  (redirect "/sitemap/1"))

(defroute ("/sitemap" :method :GET) ()
  (redirect "/sitemap/1"))

(defroute ("/sitemap/([0-9]*)" :regexp t) (&key captures)
   (let* ((pg (cond ((validation:string-is-nil-or-empty? (first captures)) 1)
                    ((first captures) (parse-integer (first captures) :junk-allowed t))
                    (t 1)))
         (page-size 1000)
         (recipes (db:get-paginated-recipes pg page-size 'updated-at :only-public t))
         (pagination (pagination:make-pagination
                      :page pg
                      :page-size page-size
                      :nb-elements (db:get-recipes-count :only-public t))))
    (when (> (gethash :page pagination) (gethash :nb-pages pagination))
      (throw-code +not-found+))
    (setf (getf (response-headers *response*) :content-type) "application/xml")
    (with-output-to-string (stream)
      (xml-emitter:with-xml-output (stream :encoding "utf-8")
        (xml-emitter:with-simple-tag
            ("urlset" nil "http://www.sitemaps.org/schemas/sitemap/0.9")
          (when (= (gethash :page pagination) 1)
            (let ((pages (db:get-all-pages :only-public t)))
              (loop :for item :in pages
                    :do (xml-emitter:with-tag ("url")
                          (xml-emitter:emit-simple-tags
                           :loc (format nil "~a/page/view/~a"
                                        (utils:build-url-root ningle:*request*)
                                        (models::page-page-id item))
                           :lastmod (format nil "~a" (mito.dao.mixin:object-updated-at item))
                           :changefreq "monthly"
                           :priority "0.5")))))
          (loop :for item :in recipes
                :do (xml-emitter:with-tag ("url")
                      (xml-emitter:emit-simple-tags
                       :loc (format nil "~a/view/recipe/~a"
                                    (utils:build-url-root ningle:*request*)
                                    (models::recipe-recipe-id item))
                       :lastmod (format nil "~a" (mito.dao.mixin:object-updated-at item))
                       :changefreq "monthly"
                       :priority "0.5")))
          (when (< (gethash :page pagination) (gethash :nb-pages pagination))
            (xml-emitter:with-tag ("sitemap")
              (xml-emitter:emit-simple-tags
               :loc (format nil "~a/sitemap/~a"
                            (utils:build-url-root ningle:*request*)
                            (1+ (gethash :page pagination)))))))))))
  
  

;; DANGER ZONE
;; =============================================================================
 
(defroute ("/danger/manage-files" :method :GET) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (render "manage-files.html"
              (append (auth:auth-user-data)
                      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data)
                        :page-files
                        ,(storage:get-file-names (uiop:directory-files +page-dir+))
                        :media-files
                        ,(storage:get-file-names (uiop:directory-files +media-dir+))
                        :thumbnail-files
                        ,(storage:get-file-names (uiop:directory-files +thumb-dir+))
                        :recipe-files
                        ,(storage:get-file-names (uiop:directory-files +recipe-dir+)))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/delete-file" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token directory file-id &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              ((find t (mapcar #'validation:string-is-nil-or-empty? `(,directory ,file-id)))
               (utils:set-alert "Directory or File Id. missing." +err+)
               (redirect "/danger/manage-files"))
              ((null (cl:probe-file
                      (storage:make-path (format nil "storage/~a/~a" directory file-id))))
               (utils:set-alert "Unable to find file" +err+)
               (redirect "/danger/manage-files"))
              (t (delete-file
                  (storage:make-path (format nil "storage/~a/~a" directory file-id)))
                 (utils:set-alert "File deleted." +suc+)
                 (redirect "/danger/manage-files"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/manage-database-files" :method :GET) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (render "manage-database-files.html"
              (append (auth:auth-user-data)
                      `(:alert ,(utils:get-and-reset-alert)
                        :system-data ,(db:get-system-data)
                        :page-files
                        ,(db:get-all-pages)
                        :media-files
                        ,(db:get-all-files)
                        :recipe-files
                        ,(db:get-all-recipes))))
      ;; Not Authorised
      (throw-code +forbidden+)))


(defroute ("/danger/delete-database-file" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token table id-key id-value &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              ((find t (mapcar
                        #'validation:string-is-nil-or-empty? `(,table ,id-key ,id-value)))
               (utils:set-alert "Database entry information is missing." +err+)
               (redirect "/danger/manage-database-files"))
              ((null (ignore-errors
                      (find (read-from-string (format nil "models:~a" table))
                            larder.web.db:*tables*)))
               (utils:set-alert "Specified table could not be found." +err+)
               (redirect "/danger/manage-database-files"))
              ((null (db:entry-exists? (read-from-string (format nil "models:~a" table))
                                       (read-from-string id-key)
                                       id-value))
               (utils:set-alert "Unable to find entry in the database" +err+)
               (redirect "/danger/manage-database-files"))
              (t (db:delete-row-entry (read-from-string (format nil "models:~a" table))
                                      (read-from-string id-key)
                                      id-value)
                 (utils:set-alert "File deleted." +suc+)
                 (redirect "/danger/manage-database-files"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/repopulate-search-db" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              (t (let ((site-settings (db:get-site-settings)))
                   (if (and (models::site_settings-search-url site-settings)
                            (models::site_settings-search-index site-settings)
                            (models::site_settings-search-api-key site-settings))
                       ;; Search integration setup.
                       (progn
                         (let ((result (meilisearch:delete-all-entries
                                        (models::site_settings-search-url site-settings)
                                        (models::site_settings-search-index site-settings)
                                        (models::site_settings-search-api-key site-settings))))
                           (when (str:contains? "Error" result)
                             (utils:set-alert result +warn+)
                             (redirect "/site-settings")))
                         (let ((result (meilisearch::submit-batch-entries
                                        (mapcar (lambda (entry) (serial:recipe-obj-to-json entry))
                                                (db:get-all-recipes :only-public t))
                                        (models::site_settings-search-url site-settings)
                                        (models::site_settings-search-index site-settings)
                                        (models::site_settings-search-api-key site-settings))))
                           (utils:set-alert result +suc+)
                           (redirect "/site-settings")))
                       ;; Search integration not setup.
                       (progn
                         (utils:set-alert "Meilisearch integration not setup." +warn+)
                         (redirect "/site-settings")))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/reset-website" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token))) (throw-code +forbidden+))
              (t (let ((site-settings (db:get-site-settings)))
                   (meilisearch:delete-all-entries
                    (models::site_settings-search-url site-settings)
                    (models::site_settings-search-index site-settings)
                    (models::site_settings-search-api-key site-settings))
                   (storage:clear-storage)
                   (db:drop-all-tables)
                   (db:ensure-tables-exist)
                   (hermetic::logout
                    ;; Successful log-out.
                    (progn (utils:flash-session-hash :id ningle:*session*)
                           (redirect "/"))
                    ;; Failed log-out
                    (progn (utils:set-alert "Failed to log out." +err+)
                           (redirect "/")))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/manage-snapshots" :method :GET) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (render "/snapshots-index.html"
	      (append (auth:auth-user-data)
		      `(:alert ,(utils:get-and-reset-alert)
		        :system-data ,(db:get-system-data)
                        :snapshots ,(uiop:subdirectories +snapshot-dir+))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/take-snapshot" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
	      (t (let* ((site-settings (db:get-site-settings))
                        (slugified-site-name
                          (utils:slugify (models::site_settings-site-name site-settings)))
                        (snapshot-timestamp (utils:create-timestamp-text))
                        (snapshot-directory
                          (format nil "~a_~a/" slugified-site-name snapshot-timestamp))
                        (snapshot-root (format nil "~a~a" +snapshot-dir+ snapshot-directory))
                        (snapshot-storage-dir
                          (format nil "~a~astorage/" +snapshot-dir+ snapshot-directory))
                        (database-dump-path
                          (format nil "~a~a_~a.sql"
                                  snapshot-root
                                  (larder.web.config:database-name)
                                  snapshot-timestamp))
                        (bash-command
                          (format nil
                                  "pg_dump -U ~a -h localhost -p 5432 -d ~a -f ~a"
                                  (larder.web.config:database-username)
                                  (larder.web.config:database-name)
                                  database-dump-path)))
                   (handler-case
                       (progn
                         #|
                         Make sure the ~/.pgpass file exists and you've ran
                         'sudo chmod 0600 ~/.pgpass' on it. Without it, this
                         section of code will always fail -- meaning no
                         snapshots. See,
                         https://www.postgresql.org/docs/current/libpq-pgpass.html
                         for more information.
                         |#
                         (ensure-directories-exist snapshot-root)
                         (utils:run-bash-command bash-command)
                         (copy-directory:copy +storage-dir+ snapshot-storage-dir))
                     (error (e)
                       (cl-fad:delete-directory-and-files snapshot-root)
                       (utils:set-alert e +err+)
                       (redirect "/danger/manage-snapshots"))))
                 (utils:set-alert "Snapshot taken." +suc+)
                 (redirect "/danger/manage-snapshots"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/delete-snapshot" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key snapshot-name authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              ((validation::string-is-nil-or-empty? snapshot-name)
               (utils:set-alert "No Snapshot Name provided." +miss+)
               (redirect "/danger/manage-snapshots"))
              ((not (cl-fad:directory-exists-p snapshot-name))
               (utils:set-alert "Cannot find selected Snapshot." +inval+)
               (redirect "/danger/manage-snapshots"))
              (t (cl-fad:delete-directory-and-files snapshot-name)
                 (utils:set-alert "Snapshot deleted." +suc+)
                 (redirect "/danger/manage-snapshots"))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/create-snapshot-download" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key snapshot-name authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
               (throw-code +forbidden+))
              ((validation::string-is-nil-or-empty? snapshot-name)
               (utils:set-alert "No Snapshot Name provided." +miss+)
               (redirect "/danger/manage-snapshots"))
              ((not (cl-fad:directory-exists-p snapshot-name))
               (utils:set-alert "Cannot find selected Snapshot." +inval+)
               (redirect "/danger/manage-snapshots"))
              ((cl:probe-file
                (format nil "~a~a" +media-dir+
                        (utils:lowercase-uuid
                         (first (last (pathname-directory snapshot-name))))))
               (utils:set-alert "Snapshot already prepared and moved to Storage Files." +inval+)
               (redirect "/danger/manage-snapshots"))
              (t (let* ((snapshot-title (first (last (pathname-directory snapshot-name))))
                        (snapshot-id (utils:lowercase-uuid snapshot-title)))
                   (zip:zip (format nil "~a~a" +media-dir+ snapshot-id)
                            (subseq snapshot-name 0 (1- (length snapshot-name)))
                            :if-exists :supersede)
                   (db:create-file snapshot-title
                                   "application/zip"
                                   (models::account-account-id (auth:get-current-user))
                                   nil
                                   nil
                                   snapshot-id)
                   (utils:set-alert "Snapshot ready for download." +suc+)
                   (redirect "/danger/manage-snapshots")))))
      ;; Not Authoried
      (throw-code +forbidden+)))

(defroute ("/danger/upload-snapshot" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authorised
      (destructuring-bind
          (&key authenticity-token &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              (t (let ((files
                         (utils:separate-files-in-web-request
                          (lack.request:request-body-parameters ningle:*request*)
                          "SNAPSHOT-FILES")))
                   (if (validation:string-is-nil-or-empty? (caddr (car files)))
                       (progn
                         (utils:set-alert "Not snapshot files provided." +miss+)
                         (redirect "/danger/manage-snapshots"))
                       (progn
                         (loop :for item :in files
                               :do (when (and (not (cl-fad:directory-exists-p
                                                    (format nil "~a~a"
                                                            +snapshot-dir+ (caddr item))))
                                              (string= "zip" (pathname-type (caddr item))))
                                     (storage:store-file
                                      (format nil "~a~a" +snapshot-dir+
                                              (utils:format-filename (caddr item)))
                                      (cdr item))
                                     (zip:unzip (format nil "~a~a" +snapshot-dir+ (caddr item))
                                                (format nil "~a~a/"
                                                        +snapshot-dir+
                                                        (pathname-name (caddr item))))
                                     (delete-file (format nil "~a~a" +snapshot-dir+
                                                          (caddr item)))))
                         (utils:set-alert "Snapshot upload complete." +suc+)
                         (redirect "/danger/manage-snapshots")))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

(defroute ("/danger/restore-snapshot" :method :POST) ()
  (hermetic:auth
      (:owner)
      ;; Authoised
      (destructuring-bind
          (&key authenticity-token snapshot-name &allow-other-keys)
          (utils:request-params (lack.request:request-body-parameters ningle:*request*))
        (cond ((not (string= authenticity-token (auth:csrf-token)))
	       (throw-code +forbidden+))
              (t (let ((ip-address
                         (handler-case
                             (utils:run-bash-command "curl https://icanhazip.com")
                           (error (e)
                             "Unable to determine the server's I.P. address."))))
                   (render "/snapshots-restore-guide.html"
	                   (append (auth:auth-user-data)
		                   `(:alert ,(utils:get-and-reset-alert)
		                     :system-data ,(db:get-system-data)
                                     :app-root
                                     ,(format nil "~a" larder.web.config:*application-root*)
                                     :ip-address ,ip-address
                                     :snapshot-name
                                     ,(first (last (pathname-directory snapshot-name)))
                                     :db-name ,(larder.web.config:database-name))))))))
      ;; Not Authorised
      (throw-code +forbidden+)))

;; ERROR PAGES
;; =============================================================================

(defmethod on-exception ((app <web>) (code (eql +not-found+)))
  (declare (ignore app))
  (merge-pathnames #P"_errors/404.html" *template-directory*))

(defmethod on-exception ((app <web>) (code (eql +forbidden+)))
  (declare (ignore app))
  (merge-pathnames #P"_errors/403-forbidden.html" *template-directory*))

#| UNSORTED ROUTES
================================================================================
Use this section to test out code and move it to its appropriate section after
you've got it working.
|#
